package uk.ac.ox.cs.goalDriven.chaseBench;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.Utilities;
import uk.ac.ox.cs.goalDriven.QA.magic.MagicProcessor;
import uk.ac.ox.cs.goalDriven.QA.magic.sips.MaximumCoverSIPs;
import uk.ac.ox.cs.goalDriven.QA.magic.sips.SIPs;
import uk.ac.ox.cs.goalDriven.QA.processors.Defun;
import uk.ac.ox.cs.goalDriven.QA.processors.Desingularization;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleRelevance;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleSubsumption;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleVerifier;
import uk.ac.ox.cs.goalDriven.QA.processors.Singularization;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class RunSOPipeline {

    public static void main(String[] args) throws Exception {
        ChaseInstance chaseInstance = new ChaseInstance();
        String scenarioNamePrefix = "";
        String scenarioName = "generated26";
        String scenarioNameSuffix = "";
        String dataVariantName = "1000";
        String queryFileName = "queries.txt";
        TRANSFORMATION transformation = TRANSFORMATION.REL;

        chaseInstance.initialize(scenarioNamePrefix, scenarioName, scenarioNameSuffix, dataVariantName, true);
        Set<Rule> queries = Utilities.loadQueries(chaseInstance, queryFileName);

        for (Rule query : queries) {
            Set<Rule> rules = new LinkedHashSet<Rule>();
            Set<Atom> facts = new LinkedHashSet<Atom>();
            // Load the rules
            if (chaseInstance.m_stTGDsFile != null)
                CommonParser.parse(chaseInstance.m_stTGDsFile, new InputCollector(rules, facts));

            rules.add(query);
            System.out.println(query);
            Atom queryAtom = query.getHeadAtom(0);

            Set<Rule> transformedRules = transformRules(chaseInstance, query, queryAtom, rules, transformation);

            // Apply the chase using RDFox
            try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(chaseInstance.usesEquality())) {
                Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(transformedRules);
                InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();

                System.out.print("Loading the source instance... ");
                System.out.flush();
                long start = System.currentTimeMillis();
                Utilities.loadCSVResources(chaseInstance, inputProcessor, predicateInfo);
                long duration = System.currentTimeMillis() - start;
                System.out.println(duration + " ms");

                inputProcessor.startProcessing();
                for (Rule rule : transformedRules)
                    inputProcessor.processRule(rule);
                inputProcessor.endProcessing();

                System.out.print("Applying " + +transformedRules.size() + " rules... ");
                System.out.flush();
                start = System.currentTimeMillis();
                jrdfoxBridge.computeMaterialization();
                duration = System.currentTimeMillis() - start;
                System.out.println(duration + " ms");

                String modelPath = scenarioName + "/" + transformation.name() + "/"
                        + query.getHeadAtom(0).getPredicate().getName() + "/" + dataVariantName;
                new File(modelPath).mkdirs();
                Integer[] entry = CommonUtilities.saveInstanceOnDisk(jrdfoxBridge, modelPath,
                        transformedRules, Arrays.asList(queryAtom.getPredicate()));
                System.out.println("Total facts " + entry[0] + "; total magic facts " + entry[1]
                        + "; total D + F facts " + entry[2] + "; useful facts " + (entry[0] - entry[1] - entry[2]));
                System.out.println();
                System.out.println();
            }
        }
    }

    public static enum TRANSFORMATION {
        MAGIC,
        REL,
        ALL
    }

    protected static Set<Rule> transformRules(ChaseInstance chaseInstance, Rule query, Atom queryAtom, Set<Rule> rules,
            TRANSFORMATION transformation) throws FileNotFoundException {
        Set<Rule> transformedRules = new LinkedHashSet<Rule>();
        InputProcessor ruleCollector = new InputCollector(transformedRules, new LinkedHashSet<Atom>());
        RuleSubsumption ruleSubsumption = null;
        if (transformation.equals(TRANSFORMATION.MAGIC) || transformation.equals(TRANSFORMATION.ALL))
            ruleSubsumption = new RuleSubsumption(new RuleVerifier(ruleCollector));
        else
            ruleSubsumption = new RuleSubsumption(ruleCollector);
        SIPs estimator = new MaximumCoverSIPs();
        InputProcessor chain = null;
        Map<Function, Integer> trueFunctions = CommonUtilities.getTrueFunctions(rules);
        String queryName = query.getHeadAtom(0).getPredicate().getName();

        Defun defun = null;
        if (chaseInstance.usesEquality())
            defun = new Defun(trueFunctions, Utilities.printer(chaseInstance, queryName, "P6"),
                    new Desingularization(Utilities.printer(chaseInstance, queryName, "P7"), ruleSubsumption));
        else
            defun = new Defun(trueFunctions, Utilities.printer(chaseInstance, queryName, "P6"), ruleSubsumption);

        if (transformation.equals(TRANSFORMATION.REL)) {
            chain = new RuleRelevance(query.getHeadAtom(0), trueFunctions, chaseInstance.usesEquality(), false,
                    Utilities.printer(chaseInstance, queryName, "P4"),
                    defun);
        } else if (transformation.equals(TRANSFORMATION.MAGIC)) {
            chain = new MagicProcessor(queryAtom, estimator, trueFunctions, Utilities
                    .printer(chaseInstance, queryName, "P5"),
                    defun);
        } else if (transformation.equals(TRANSFORMATION.ALL)) {
            chain = new RuleRelevance(query.getHeadAtom(0), trueFunctions, chaseInstance.usesEquality(), false,
                    Utilities
                            .printer(chaseInstance, queryName, "P4"),
                    new MagicProcessor(queryAtom, estimator, trueFunctions, Utilities
                            .printer(chaseInstance, queryName, "P5"),
                            defun));
        }

        if (chaseInstance.usesEquality())
            chain = new Singularization(queryAtom.getPredicate(),
                    Utilities.printer(chaseInstance, queryName, "P2"), chain);

        chain.startProcessing();
        for (Rule inputRule : rules)
            chain.processRule(inputRule);
        chain.endProcessing();
        transformedRules.addAll(Utilities.createAuxiliarySOAxiomsForBodyFunctions(defun.getBodyPredicates()));
        transformedRules.addAll(CommonUtilities.getDomainAxioms(transformedRules));

        CommonFormatProcessor P8 = Utilities.printer(chaseInstance, queryName, "P8");
        P8.startProcessing();
        for (Rule inputRule : transformedRules)
            P8.processRule(inputRule);
        P8.endProcessing();

        return transformedRules;
    }
}
