package uk.ac.ox.cs.goalDriven.QA.magic;

import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import uk.ac.ox.cs.goalDriven.QA.magic.sips.SIPs;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

import java.util.Set;

public class MagicProcessor implements InputProcessor {
    protected final InputProcessor m_inputProcessor;
    protected final SIPs m_strategy;
    protected Map<Predicate, Set<Rule>> m_indexedRules;
    protected final Atom m_query;
    protected Map<Predicate, PredicateInfo> m_predicateInfo;
    protected Map<Function, Integer> m_trueFunctions;
    protected Set<Rule> m_rules = new LinkedHashSet<>();
    protected Set<Rule> m_magicRules = new LinkedHashSet<>();
    protected final CommonFormatProcessor m_writter;

    public MagicProcessor(Atom query,
            SIPs strategy,
            Map<Function, Integer> trueFunctions,
            CommonFormatProcessor writter,
            InputProcessor inputProcessor) {
        m_strategy = strategy;
        m_query = query;
        m_inputProcessor = inputProcessor;
        m_trueFunctions = trueFunctions;
        m_writter = writter;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
        m_writter.startProcessing();
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
        m_writter.processFact(predicate, argumentLexicalForms);
    }

    public void processRule(Rule rule) {
        if (rule.getHeadAtoms().length > 1)
            throw new java.lang.RuntimeException("Non datalog rules are not supported");
        m_rules.add(rule);
    }

    public void endProcessing() {

        Map<Function, Integer> trueFunctions = CommonUtilities.getTrueFunctions(m_rules);
        m_trueFunctions.keySet().retainAll(trueFunctions.keySet());

        Set<Rule> DAxioms = CommonUtilities.getDAxiomatization(m_rules);
        m_indexedRules = CommonUtilities.createRuleIndex(m_rules);
        m_indexedRules.put(CommonUtilities.D, DAxioms);

        // Add the magic rule for the query
        Boolean[] queryAdornment = new Boolean[m_query.getNumberOfArguments()];
        for (int index = 0; index < m_query.getNumberOfArguments(); ++index) {
            if (m_query.getArgument(index) instanceof Constant)
                queryAdornment[index] = true;
            else
                queryAdornment[index] = false;
        }

        Atom magicAtom = MagicUtilities.computeMagicAtom(m_query, queryAdornment);
        m_inputProcessor.processRule(Rule.create(magicAtom));
        m_writter.processRule(Rule.create(magicAtom));

        // Compute the sips for rules. It removes subsumed sips
        Map<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> adornmentsForRules = MagicUtilities
                .computeSIPSAndRemoveSubsumedOnes(m_query, m_indexedRules, m_strategy, m_trueFunctions);

        Boolean existsRuleWithEquality = false;
        for (Entry<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> entry : adornmentsForRules.entrySet()) {
            Entry<Rule, Boolean[]> ruleAndHeadAdornment = entry.getKey();
            Entry<Atom[], Boolean[][]> bodyOrderingAndBodyAdornment = entry.getValue();
            process(ruleAndHeadAdornment.getKey(), ruleAndHeadAdornment.getValue(),
                    bodyOrderingAndBodyAdornment.getKey(), bodyOrderingAndBodyAdornment.getValue());
            if (ruleAndHeadAdornment.getKey().getHeadAtom(0).getPredicate().equals(Predicate.EQUALS))
                existsRuleWithEquality = true;
        }

        // add follow-up magic axioms
        if (existsRuleWithEquality && m_trueFunctions.size() > 0) {
            for (Rule axiom : MagicUtilities.addFollowUpEqualityAxioms(m_trueFunctions)) {
                m_inputProcessor.processRule(axiom);
                m_writter.processRule(axiom);
            }
        }

        m_inputProcessor.endProcessing();
        m_writter.endProcessing();
    }

    private void process(Rule rule, Boolean[] headAdornment, Atom[] bodyOrdering, Boolean[][] bodyAdornment) {

        Atom magicHeadAtom = MagicUtilities.computeMagicAtom(rule.getHeadAtom(0), headAdornment);
        if (!rule.getHeadAtom(0).getPredicate().equals(CommonUtilities.D)) {
            Atom[] modifiedRuleBody = new Atom[1 + bodyOrdering.length];
            modifiedRuleBody[0] = magicHeadAtom;
            System.arraycopy(bodyOrdering, 0, modifiedRuleBody, 1, bodyOrdering.length);
            m_inputProcessor.processRule(Rule.create(rule.getHeadAtoms(), modifiedRuleBody));
            m_writter.processRule(Rule.create(rule.getHeadAtoms(), modifiedRuleBody));
        }

        for (int bodyAtomIndex = 0; bodyAtomIndex < bodyOrdering.length; ++bodyAtomIndex) {
            Atom bodyAtom = bodyOrdering[bodyAtomIndex];
            if (m_indexedRules.containsKey(bodyAtom.getPredicate())) {
                Atom magicBodyAtom = MagicUtilities.computeMagicAtom(bodyAtom, bodyAdornment[bodyAtomIndex]);
                Atom[] magicRuleHead = new Atom[] { magicBodyAtom };
                Atom[] magicRuleBody = new Atom[bodyAtomIndex + 1];
                magicRuleBody[0] = magicHeadAtom;
                System.arraycopy(bodyOrdering, 0, magicRuleBody, 1, bodyAtomIndex);
                Rule magicRule = Rule.create(magicRuleHead, magicRuleBody);
                if (!Arrays.equals(magicRuleHead, magicRuleBody) && !m_magicRules.contains(magicRule)) {
                    m_magicRules.add(magicRule);
                    m_inputProcessor.processRule(magicRule);
                    m_writter.processRule(magicRule);
                }
            }
        }
    }
}