package uk.ac.ox.cs.goalDriven.QA.magic;

import java.util.AbstractMap;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import uk.ac.ox.cs.goalDriven.QA.magic.sips.SIPs;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

import java.util.Set;

public class MagicUtilities {

    public static final String magicPrefix = "m_";

    public static Map<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> computeSIPSAndRemoveSubsumedOnes(Atom query,
            Map<Predicate, Set<Rule>> indexedRules, SIPs strategy, Map<Function, Integer> trueFunctions) {
        Map<Binding, Binding> adornmentReplacements = new LinkedHashMap<>();
        Map<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> adornmentsForRules = MagicUtilities
                .computeSIPS(query, indexedRules, strategy, trueFunctions, adornmentReplacements);
        // Find the subsumed adornments and revise the previous sips
        Set<Entry<Rule, Boolean[]>> ruleAndHeadAdornments = adornmentsForRules.keySet();
        int index = 0;
        Binding[] bindings = new Binding[ruleAndHeadAdornments.size()];
        for (Entry<Rule, Boolean[]> entry : ruleAndHeadAdornments)
            bindings[index++] = Binding.create(entry.getKey().getHeadAtom(0).getPredicate(), entry.getValue());
        adornmentReplacements = MagicUtilities.computeSubsumedAdornments(bindings);
        // Revise the sips
        adornmentsForRules = MagicUtilities.computeSIPS(query, indexedRules, strategy, trueFunctions,
                adornmentReplacements);
        return adornmentsForRules;
    }

    public static Map<Binding, Binding> computeSubsumedAdornments(Binding[] patterns) {
        Map<Binding, Binding> adornmentReplacements = new LinkedHashMap<>();
        Graph<Binding> graph = new Graph<Binding>();
        for (int i = 0; i < patterns.length; ++i) {
            for (int j = 0; j < patterns.length; ++j) {
                if (patterns[j].subsumes(patterns[i]))
                    graph.addEdge(patterns[i], patterns[j]);
            }
        }
        for (int i = 0; i < patterns.length; ++i) {
            Binding sync = graph.findSync(patterns[i]);
            if (!patterns[i].equals(sync))
                adornmentReplacements.put(patterns[i], sync);
        }
        return adornmentReplacements;
    }

    public static Map<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> computeSIPS(Atom query,
            Map<Predicate, Set<Rule>> indexedRules, SIPs strategy,
            Map<Function, Integer> trueFunctions,
            Map<Binding, Binding> adornmentReplacements) {

        // Each rule is associated with a head adornment, a body ordering, and a list of
        // adornments.
        // Rule/head adornment, body ordering/adornments
        Map<Entry<Rule, Boolean[]>, Entry<Atom[], Boolean[][]>> sips = new LinkedHashMap<>();
        Set<Binding> freshBinding = new LinkedHashSet<>();
        Set<Binding> pastBindings = new LinkedHashSet<>();
        Boolean firstTime = true;
        // Compute the adornment for the query
        Boolean[] queryAdornment = new Boolean[query.getNumberOfArguments()];
        for (int index = 0; index < query.getNumberOfArguments(); ++index) {
            if (query.getArgument(index) instanceof Constant)
                queryAdornment[index] = true;
            else
                queryAdornment[index] = false;
        }
        Binding goalBinding = Binding.create(query.getPredicate(), queryAdornment);
        freshBinding.add(goalBinding);
        while (!freshBinding.isEmpty()) {
            Binding binding = freshBinding.iterator().next();
            freshBinding.remove(binding);
            pastBindings.add(binding);
            Boolean[] headAdornment = binding.getAdornment();
            Collection<Rule> rules = indexedRules.get(binding.m_predicate);
            if (rules != null)
                for (Rule rule : rules) {
                    Term[] terms = MagicUtilities.computeBoundArguments(rule.getHeadAtom(0).getArguments(),
                            headAdornment);
                    Set<Variable> variables = new LinkedHashSet<>();
                    for (Term term : terms)
                        term.getFreeVariables(variables);

                    Entry<Atom[], Boolean[][]> bodyOrderingAndBodyAdornments = strategy.sips(rule,
                            variables, adornmentReplacements);

                    // The newly created adornments should be processed next
                    Atom[] bodyOrdering = bodyOrderingAndBodyAdornments.getKey();
                    Boolean[][] bodyAdornments = bodyOrderingAndBodyAdornments.getValue();
                    for (int index = 0; index < bodyOrdering.length; ++index) {
                        Atom bodyAtom = bodyOrdering[index];
                        Boolean[] adornment = bodyAdornments[index];
                        if (bodyAtom.getPredicate().equals(Predicate.EQUALS) && firstTime) {
                            Binding p1 = Binding.create(bodyAtom.getPredicate(), new Boolean[] { true, false });
                            Binding p2 = Binding.create(bodyAtom.getPredicate(), new Boolean[] { false, true });
                            freshBinding.add(p1);
                            freshBinding.add(p2);
                            firstTime = false;
                            if (trueFunctions.size() > 0)
                                freshBinding.add(Binding.create(CommonUtilities.D, new Boolean[] { true }));
                        } else {
                            Binding p = Binding.create(bodyAtom.getPredicate(), adornment);
                            if (!pastBindings.contains(p))
                                freshBinding.add(p);
                        }
                    }
                    SimpleEntry<Rule, Boolean[]> e = new AbstractMap.SimpleEntry<Rule, Boolean[]>(rule, headAdornment);
                    if (!sips.containsKey(e))
                        sips.put(e, bodyOrderingAndBodyAdornments);
                }
        }
        return sips;
    }

    public static Term[] computeBoundArguments(Term[] terms, Boolean[] adornment) {
        List<Term> boundTerms = new ArrayList<>();
        for (int index = 0; index < adornment.length; ++index) {
            if (adornment[index])
                boundTerms.add(terms[index]);
        }
        return boundTerms.toArray(new Term[boundTerms.size()]);
    }

    public static Atom computeMagicAtom(Atom atom, Boolean[] adornment) {
        Term[] arguments = MagicUtilities.computeBoundArguments(atom.getArguments(), adornment);
        Predicate magicPredicate = MagicUtilities.convertBindingToMagicPredicate(atom.getPredicate(), adornment);
        return Atom.create(magicPredicate, arguments);
    }

    public static Predicate convertBindingToMagicPredicate(Predicate predicate, Boolean[] adornment) {
        String ad = "";
        if (predicate.equals(Predicate.EQUALS))
            ad = "b";
        else {
            for (int adornmentIndex = 0; adornmentIndex < adornment.length; ++adornmentIndex) {
                if (adornment[adornmentIndex])
                    ad += "b";
                else
                    ad += "f";
            }
        }
        return Predicate.create(magicPrefix + predicate.getName() + "_" + ad);
    }

    public static Set<Rule> addFollowUpEqualityAxioms(Map<Function, Integer> trueFunctions) {
        Set<Rule> axioms = new LinkedHashSet<>();
        Term x1 = Variable.create("X1");
        Term x2 = Variable.create("X2");
        Atom X1X2 = Atom.create(Predicate.EQUALS, x1, x2);
        Predicate magicEQPredicate = MagicUtilities.convertBindingToMagicPredicate(Predicate.EQUALS,
                new Boolean[] { true, false });
        axioms.add(Rule.create(Atom.create(magicEQPredicate, x2),
                Atom.create(magicEQPredicate, x1), X1X2));

        for (Entry<Function, Integer> e : trueFunctions.entrySet()) {
            Function function = e.getKey();
            Integer arity = e.getValue();
            Term[] terms = new Term[arity];
            for (int i = 0; i < arity; ++i)
                terms[i] = Variable.create("X" + i);
            FunctionTerm fx1xn = FunctionTerm.create(function, terms);
            Atom magicEQfx1xnAtom = Atom.create(magicEQPredicate, fx1xn);
            Predicate magicDPredicate = MagicUtilities.convertBindingToMagicPredicate(CommonUtilities.D,
                    new Boolean[] { true });
            for (int i = 0; i < arity; ++i) {
                Variable xi = Variable.create("X" + i);
                Variable xip = Variable.create("X" + i + "p");
                axioms.add(Rule.create(Atom.create(magicDPredicate, xi),
                        magicEQfx1xnAtom));
                axioms.add(Rule.create(Atom.create(magicEQPredicate, xi),
                        magicEQfx1xnAtom, Atom.create(CommonUtilities.D, xi)));
                axioms.add(Rule.create(Atom.create(magicDPredicate, xip),
                        magicEQfx1xnAtom, Atom.create(CommonUtilities.D, xi),
                        Atom.create(Predicate.EQUALS, xi, xip)));
            }
        }
        return axioms;
    }

}
