package uk.ac.ox.cs.goalDriven.QA.magic.sips;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.magic.MagicUtilities;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

public class SIPSUtilities {

    public static Atom[] greedyRuleOrderingLegacy(Rule rule, Boolean[] adornment) {
        // Compute all bound variables from the head
        Term[] terms = MagicUtilities.computeBoundArguments(rule.getHeadAtom(0).getArguments(), adornment);
        Set<Variable> boundVariables = new LinkedHashSet<>();
        for (Term term : terms)
            term.getFreeVariables(boundVariables);
        return greedyRuleOrdering(rule.getBodyAtoms(), boundVariables);
    }

    public static Atom[] greedyRuleOrdering(Atom[] bodyAtoms, Set<Variable> boundVariablesInit) {
        Set<Variable> boundVariables = new LinkedHashSet<>(boundVariablesInit);
        // At position j put the atom that gets the maximum inputs from its ancestors
        List<Atom> newBody = new ArrayList<>();
        List<Atom> oldBody = new LinkedList<>(Arrays.asList(bodyAtoms));
        while (true) {
            Atom atom = null;
            if (boundVariables.isEmpty())
                atom = getRelationalAtomWithMostConstantsOrMostAtomsToCover(oldBody);
            else
                atom = getAtomWithMostCoveredArguments(newBody, oldBody, boundVariables);
            atom.getFreeVariables(boundVariables);
            newBody.add(atom);
            oldBody.remove(atom);
            if (oldBody.isEmpty())
                break;
        }
        return newBody.toArray(new Atom[newBody.size()]);
    }

    private static Atom getAtomWithMostCoveredArguments(List<Atom> precedingAtoms, List<Atom> bodyAtoms,
            Set<Variable> boundArguments) {
        Integer maxNumberOfCoveredArguments = Integer.MIN_VALUE;
        Atom atomWithMaxNumberOfNumberOfCoveredArguments = null;
        for (int bodyAtomIndex = 0; bodyAtomIndex < bodyAtoms.size(); ++bodyAtomIndex) {
            Atom bodyAtom = bodyAtoms.get(bodyAtomIndex);
            Integer numberOfCoveredArguments = 0;
            for (int argumentIndex = 0; argumentIndex < bodyAtom.getNumberOfArguments(); ++argumentIndex) {
                Term argument = bodyAtom.getArgument(argumentIndex);
                if (argument instanceof Constant)
                    ++numberOfCoveredArguments;
                else {
                    Set<Variable> freeVariables = new LinkedHashSet<>();
                    argument.getFreeVariables(freeVariables);
                    if (boundArguments.containsAll(freeVariables))
                        ++numberOfCoveredArguments;
                }
            }
            if (numberOfCoveredArguments > maxNumberOfCoveredArguments && equalityIsGuarded(precedingAtoms, bodyAtom)) {
                atomWithMaxNumberOfNumberOfCoveredArguments = bodyAtom;
                maxNumberOfCoveredArguments = numberOfCoveredArguments;
            }
        }
        return atomWithMaxNumberOfNumberOfCoveredArguments;
    }

    private static Boolean equalityIsGuarded(List<Atom> precedingAtoms, Atom atom) {
        if (!atom.getPredicate().equals(Predicate.EQUALS))
            return true;
        // equality atoms where no variables occur are guarded
        Set<Variable> variables = new LinkedHashSet<>();
        atom.getFreeVariables(variables);
        if (variables.size() == 0)
            return true;
        // Get all variables occurring in relational atoms only
        variables = new LinkedHashSet<>();
        for (int index = 0; index < precedingAtoms.size(); ++index) {
            Atom precedingAtom = precedingAtoms.get(index);
            if (!precedingAtom.getPredicate().equals(Predicate.EQUALS))
                precedingAtom.getFreeVariables(variables);
        }
        for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
            List<Variable> freeVariables = new ArrayList<>();
            atom.getArgument(argumentIndex).getFreeVariables(freeVariables);
            if (freeVariables.size() > 0 && variables.containsAll(freeVariables))
                return true;
        }
        return false;
    }

    public static Atom getRelationalAtomWithMostConstantsOrMostAtomsToCover(List<Atom> bodyAtoms) {
        Integer maxNumberOfConstants = Integer.MIN_VALUE;
        Atom atomWithMaxNumberOfInputs = null;
        for (int bodyAtomIndex = 0; bodyAtomIndex < bodyAtoms.size(); ++bodyAtomIndex) {
            Set<Constant> constants = new LinkedHashSet<>();
            Atom bodyAtom = bodyAtoms.get(bodyAtomIndex);
            bodyAtom.getConstants(constants);
            if (constants.size() > maxNumberOfConstants && !bodyAtom.getPredicate().equals(Predicate.EQUALS)) {
                atomWithMaxNumberOfInputs = bodyAtom;
                maxNumberOfConstants = constants.size();
            }
        }
        if (maxNumberOfConstants > 0)
            return atomWithMaxNumberOfInputs;
        else {
            Set<Variable> allVariables = new LinkedHashSet<>();
            Atom atomCoveringMostAtoms = null;
            Integer maxNumberOfCoveringVariables = Integer.MIN_VALUE;
            for (int bodyAtomIndex = 0; bodyAtomIndex < bodyAtoms.size(); ++bodyAtomIndex)
                bodyAtoms.get(bodyAtomIndex).getFreeVariables(allVariables);
            for (int bodyAtomIndex = 0; bodyAtomIndex < bodyAtoms.size(); ++bodyAtomIndex) {
                Atom bodyAtom = bodyAtoms.get(bodyAtomIndex);
                Set<Variable> intersection = new LinkedHashSet<>();
                bodyAtom.getFreeVariables(intersection);
                intersection.retainAll(allVariables);
                if (intersection.size() > maxNumberOfCoveringVariables
                        && !bodyAtom.getPredicate().equals(Predicate.EQUALS)) {
                    maxNumberOfCoveringVariables = intersection.size();
                    atomCoveringMostAtoms = bodyAtom;
                }
            }
            return atomCoveringMostAtoms;
        }
    }

}
