package uk.ac.ox.cs.goalDriven.QA.processors;

import static org.junit.Assert.assertTrue;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class Defun implements InputProcessor {
    protected final InputProcessor m_inputProcessor;
    protected final Set<Rule> m_rules = new LinkedHashSet<>();
    protected final Map<Function, Integer> m_trueFunctions;
    protected final Map<Predicate, Integer> m_bodyPredicates = new LinkedHashMap<>();
    protected final Map<Function, Integer> m_bodyFunctions = new LinkedHashMap<>();
    protected final Map<Constant, Constant> m_constantsMap = new LinkedHashMap<>();
    protected int m_constantIndex = 0;
    protected final CommonFormatProcessor m_writter;

    public Defun(Map<Function, Integer> trueFunctions, CommonFormatProcessor writter, InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
        m_trueFunctions = trueFunctions;
        m_writter = writter;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
        m_writter.startProcessing();
    }

    public void processRule(Rule rule) {
        assertTrue(rule.getNumberOfHeadAtoms() == 1);
        m_rules.add(rule);
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
        m_writter.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {

        Set<Rule> newRules = new LinkedHashSet<>();
        for (Rule rule : m_rules) {
            List<Atom> newBody = new ArrayList<Atom>();
            for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
                Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
                Term[] newArguments = bodyAtom.getArguments();
                for (int argumentIndex = 0; argumentIndex < newArguments.length; ++argumentIndex) {
                    if (newArguments[argumentIndex] instanceof FunctionTerm) {
                        FunctionTerm argument = (FunctionTerm) newArguments[argumentIndex];
                        AbstractMap.SimpleEntry<Variable, Predicate> entry = createFreshVariableAndPredicate(argument);
                        Variable freshVariable = entry.getKey();
                        newArguments[argumentIndex] = freshVariable;
                        Predicate freshPredicate = entry.getValue();
                        List<Term> freshAtomTerms = new ArrayList<>();
                        for (int index = 0; index < argument.getNumberOfArguments(); ++index)
                            freshAtomTerms.add(argument.getArgument(index));
                        freshAtomTerms.add(freshVariable);
                        m_bodyPredicates.put(freshPredicate, freshAtomTerms.size());
                        m_bodyFunctions.put(argument.getFunction(), argument.getNumberOfArguments());
                        Atom freshAtom = Atom.create(freshPredicate,
                                freshAtomTerms.toArray(new Term[freshAtomTerms.size()]));
                        if (!newBody.contains(freshAtom))
                            newBody.add(freshAtom);
                    } else if (newArguments[argumentIndex] instanceof Constant) {
                        Constant argument = (Constant) newArguments[argumentIndex];
                        AbstractMap.SimpleEntry<Variable, Predicate> entry = createFreshVariableAndPredicate(argument);
                        Variable freshVariable = entry.getKey();
                        newArguments[argumentIndex] = freshVariable;
                        Predicate freshPredicate = entry.getValue();
                        Atom freshAtom = Atom.create(freshPredicate, freshVariable);
                        if (!newBody.contains(freshAtom)) {
                            newBody.add(freshAtom);
                            m_inputProcessor.processRule(Rule.create(Atom.create(freshPredicate, argument)));
                            m_writter.processRule(Rule.create(Atom.create(freshPredicate, argument)));
                        }
                    }
                }
                newBody.add(Atom.create(bodyAtom.getPredicate(), newArguments));
            }
            newRules.add(Rule.create(rule.getHeadAtom(0), newBody.toArray(new Atom[newBody.size()])));
        }

        for (Rule rule : newRules) {
            Atom headAtom = rule.getHeadAtom(0);
            for (Entry<Function, Integer> entry : m_bodyFunctions.entrySet()) {
                Function f = entry.getKey();
                List<FunctionTerm> matchingFunctionTerms = matchingFunctionTerms(headAtom, f);
                for (FunctionTerm t : matchingFunctionTerms) {
                    Predicate freshPredicate = Predicate.create("F_" + f.getName());
                    List<Term> freshAtomTerms = new ArrayList<>();
                    for (int index = 0; index < t.getNumberOfArguments(); ++index)
                        freshAtomTerms.add(t.getArgument(index));
                    freshAtomTerms.add(t);
                    Atom freshAtom = Atom.create(freshPredicate,
                            freshAtomTerms.toArray(new Term[freshAtomTerms.size()]));
                    m_inputProcessor.processRule(Rule.create(freshAtom, rule.getBodyAtoms()));
                    m_writter.processRule(Rule.create(freshAtom, rule.getBodyAtoms()));
                }
            }
            m_inputProcessor.processRule(rule);
            m_writter.processRule(rule);
        }

        // Follow-up steps related to SO TGDs
        Map<Function, Integer> trueFunctions = CommonUtilities.getTrueFunctions(newRules);
        m_trueFunctions.keySet().retainAll(trueFunctions.keySet());

        for (Entry<Function, Integer> entry : m_trueFunctions.entrySet()) {
            Function function = entry.getKey();
            Predicate predicate = Predicate.create("F_" + function.getName());
            int arity = entry.getValue();
            Term[] variables = new Term[arity];
            Term[] variables_p = new Term[arity];
            Term[] terms = new Term[arity + 1];
            List<Atom> bodyAtoms = new ArrayList<>();
            for (int i = 0; i < arity; ++i) {
                variables[i] = Variable.create("X" + i);
                variables_p[i] = Variable.create("Xp" + i);
                terms[i] = Variable.create("X" + i);
                bodyAtoms.add(Atom.create(CommonUtilities.D, variables[i]));
                bodyAtoms.add(Atom.create(Predicate.EQUALS, variables[i], variables_p[i]));
                bodyAtoms.add(Atom.create(CommonUtilities.D, variables_p[i]));
            }
            terms[arity] = FunctionTerm.create(function, variables);
            Atom head = Atom.create(predicate, terms);
            m_inputProcessor.processRule(Rule.create(head, bodyAtoms.toArray(new Atom[0])));
            m_writter.processRule(Rule.create(head, bodyAtoms.toArray(new Atom[0])));
        }

        m_inputProcessor.endProcessing();
        m_writter.endProcessing();

    }

    public Map<Predicate, Integer> getBodyPredicates() {
        return this.m_bodyPredicates;
    }

    private List<FunctionTerm> matchingFunctionTerms(Atom headAtom, Function f) {
        List<FunctionTerm> terms = new ArrayList<>();
        for (int index = 0; index < headAtom.getNumberOfArguments(); ++index) {
            Term t = headAtom.getArgument(index);
            if (t instanceof FunctionTerm && ((FunctionTerm) t).getFunction().equals(f))
                terms.add((FunctionTerm) t);
        }
        return terms;
    }

    private AbstractMap.SimpleEntry<Variable, Predicate> createFreshVariableAndPredicate(FunctionTerm term) {
        String name = "z_";
        name += term.getFunction().getName().toString();
        if (term.getNumberOfArguments() > 0) {
            name += '_';
            for (int index = 0; index < term.getNumberOfArguments(); index++) {
                Term argument = term.getArgument(index);
                if (index > 0)
                    name += '_';
                name += argument.toString().replaceAll("\\?", "");
            }
        }
        Variable freshVariable = Variable.create(name);
        Predicate freshPredicate = Predicate.create("F_" + term.getFunction().getName());
        return new AbstractMap.SimpleEntry<Variable, Predicate>(freshVariable, freshPredicate);
    }

    private AbstractMap.SimpleEntry<Variable, Predicate> createFreshVariableAndPredicate(Constant constant) {
        Constant newConstant;
        if (!m_constantsMap.containsKey(constant)) {
            newConstant = Constant.create("dummy_" + m_constantIndex++);
            m_constantsMap.put(constant, newConstant);
        } else
            newConstant = m_constantsMap.get(constant);

        Variable freshVariable = Variable.create("z_" + newConstant.toString());
        Predicate freshPredicate = Predicate.create("F_" + newConstant.toString());
        return new AbstractMap.SimpleEntry<Variable, Predicate>(freshVariable, freshPredicate);
    }
}
