package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class RuleRelevance implements InputProcessor {

    protected InputProcessor m_inputProcessor;
    protected final Boolean m_usesEquality;
    protected final Atom m_query;
    protected Set<Rule> m_rules;
    protected final Boolean m_doFastRelevanceAnalysis;
    protected Map<Predicate, PredicateInfo> m_predicateInfo;
    protected final Map<Function, Integer> m_trueFunctions;
    protected final CommonFormatProcessor m_writter;

    public RuleRelevance(Atom query,
            Map<Function, Integer> trueFunctions, Boolean usesEquality, Boolean doFastRelevanceAnalysis,
            CommonFormatProcessor writter,
            InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
        m_usesEquality = usesEquality;
        m_query = query;
        m_trueFunctions = trueFunctions;
        m_rules = new LinkedHashSet<>();
        m_doFastRelevanceAnalysis = doFastRelevanceAnalysis;
        m_writter = writter;
    }

    @Override
    public void startProcessing() {
        m_inputProcessor.startProcessing();
        m_writter.startProcessing();
    }

    @Override
    public void processRule(Rule rule) {
        m_rules.add(rule);
    }

    @Override
    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
        m_writter.processFact(predicate, argumentLexicalForms);
    }

    @Override
    public void endProcessing() {
        m_predicateInfo = CommonUtilities.getPredicateInfo(m_rules);

        try {
            try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(false)) {

                InputProcessor jrdfoxProcessor = jrdfoxBridge.newInputProcessor();
                jrdfoxProcessor.startProcessing();
                Map<Predicate, List<String>> criticalInstance = RuleRelevanceUtilities
                        .createCriticalInstance(m_predicateInfo);
                for (Predicate predicate : criticalInstance.keySet())
                    jrdfoxProcessor.processFact(predicate, criticalInstance.get(predicate));
                jrdfoxProcessor.endProcessing();

                // The input rules are singularized and skolemized
                // If we do the fast relevance analysis, firstly, we should replace function
                // terms with nulls and modify appropriately the input processor
                Set<Rule> singularisedRulesForForwardChaining = new LinkedHashSet<Rule>();
                InputProcessor chain = new InputCollector(singularisedRulesForForwardChaining,
                        new LinkedHashSet<Atom>());

                if (m_doFastRelevanceAnalysis)
                    chain = new ReplaceFunctionTermsWithNulls(chain);
                chain.startProcessing();
                for (Rule inputRule : m_rules)
                    chain.processRule(inputRule);
                chain.endProcessing();
                if (m_doFastRelevanceAnalysis)
                    m_inputProcessor = new ReplaceNullsWithFunctionTerms(
                            ((ReplaceFunctionTermsWithNulls) chain).getAtomsMap(), m_inputProcessor);

                // Replace constants in the rule bodies with the critical constant;
                Set<Rule> singularisedRulesForForwardChainingWithCriticalConstants = RuleRelevanceUtilities
                        .replaceBodyConstantsWithCriticalConstants(
                                singularisedRulesForForwardChaining);

                // Axiomatise equality
                Set<Rule> DAxioms = new LinkedHashSet<>();
                Set<Rule> FRAxioms = new LinkedHashSet<>();

                if (m_usesEquality) {
                    DAxioms = CommonUtilities
                            .getDAxiomatization(singularisedRulesForForwardChainingWithCriticalConstants);
                    FRAxioms = CommonUtilities.getFRAxiomatization(m_trueFunctions);
                    singularisedRulesForForwardChainingWithCriticalConstants.addAll(DAxioms);
                    singularisedRulesForForwardChainingWithCriticalConstants.add(CommonUtilities.R);
                    singularisedRulesForForwardChainingWithCriticalConstants.add(CommonUtilities.symmetry);
                    singularisedRulesForForwardChainingWithCriticalConstants.add(CommonUtilities.transitivity);
                    singularisedRulesForForwardChainingWithCriticalConstants.addAll(FRAxioms);
                }
                // Chase the critical instance
                System.out.print("Loading rules... ");
                System.out.flush();
                long startLoadRules = System.currentTimeMillis();
                jrdfoxProcessor.startProcessing();
                for (Rule rule : singularisedRulesForForwardChainingWithCriticalConstants)
                    jrdfoxProcessor.processRule(rule);
                jrdfoxProcessor.endProcessing();
                long durationLoadRules = System.currentTimeMillis() - startLoadRules;
                System.out.println(durationLoadRules + " ms");

                System.out.print("Chasing the critical instance... ");
                System.out.flush();
                long startQueryMat = System.currentTimeMillis();
                jrdfoxBridge.computeMaterialization();
                long durationQueryMat = System.currentTimeMillis() - startQueryMat;
                System.out.println(durationQueryMat + " ms");

                Set<Rule> singularisedRulesForBackwardChaining = new LinkedHashSet<>();
                singularisedRulesForBackwardChaining.addAll(singularisedRulesForForwardChaining);
                if (m_usesEquality) {
                    singularisedRulesForBackwardChaining.addAll(DAxioms);
                    singularisedRulesForBackwardChaining.add(CommonUtilities.symmetry);
                    singularisedRulesForBackwardChaining.add(CommonUtilities.transitivity);
                    singularisedRulesForBackwardChaining.addAll(FRAxioms);
                }

                Map<Predicate, Set<Rule>> indexedSingularisedRulesForBackwardChaining = CommonUtilities
                        .createRuleIndex(singularisedRulesForBackwardChaining);

                // Find relevant rules
                Set<Atom> seenFacts = new LinkedHashSet<>();
                Set<Atom> toProcessFacts = new LinkedHashSet<>();
                Map<Rule, Set<Integer>> relevantRules = new LinkedHashMap<>();

                // Create the goal ground fact
                int queryArity = m_predicateInfo.get(m_query.getPredicate()).getArity();
                Constant[] arguments = new Constant[queryArity];
                for (int column = 0; column < queryArity; ++column)
                    arguments[column] = CommonUtilities.critical;
                seenFacts.add(Atom.create(m_query.getPredicate(), arguments));
                toProcessFacts.add(Atom.create(m_query.getPredicate(), arguments));

                System.out.print("Answering queries for rule relevance... ");
                System.out.flush();
                long start = System.currentTimeMillis();

                while (!toProcessFacts.isEmpty()) {

                    Iterator<Atom> iterator = toProcessFacts.iterator();
                    Atom fact = iterator.next();
                    iterator.remove();
                    Predicate predicate = fact.getPredicate();
                    Set<Rule> rules = indexedSingularisedRulesForBackwardChaining.get(predicate);
                    if (rules != null) {
                        for (Rule rule : rules) {
                            Rule ruleWithoutConstants = RuleRelevanceUtilities
                                    .replaceBodyConstantsWithCriticalConstant(rule);

                            Set<Map<Variable, Term>> substitutions = RuleRelevanceUtilities
                                    .computeAllValidSubstitutions(
                                            jrdfoxBridge,
                                            ruleWithoutConstants, fact);

                            for (Map<Variable, Term> nu : substitutions) {
                                if (!relevantRules.containsKey(rule) &&
                                        !rule.equals(CommonUtilities.transitivity)
                                        && !rule.equals(CommonUtilities.symmetry)
                                        && !DAxioms.contains(rule)
                                        && !FRAxioms.contains(rule)) {
                                    relevantRules.put(rule, new LinkedHashSet<>());
                                }

                                Atom[] instantiatedBodyAtoms = RuleRelevanceUtilities
                                        .applySubstitution(ruleWithoutConstants.getBodyAtoms(), nu);

                                for (int instantiatedAtomIndex = 0; instantiatedAtomIndex < instantiatedBodyAtoms.length; ++instantiatedAtomIndex) {
                                    Atom G = (Atom) instantiatedBodyAtoms[instantiatedAtomIndex];
                                    // Find the important equalities
                                    if (!RuleRelevanceUtilities.isTrivialEquality(G)) {
                                        if (!seenFacts.contains(G)) {
                                            seenFacts.add(G);
                                            toProcessFacts.add(G);
                                        }

                                        if (G.getPredicate().equals(Predicate.EQUALS) && !DAxioms.contains(rule)
                                                && !rule.equals(CommonUtilities.symmetry)
                                                && !rule.equals(CommonUtilities.transitivity)
                                                && !FRAxioms.contains(rule))
                                            relevantRules.get(rule).add(instantiatedAtomIndex);
                                    }
                                }
                            }
                        }
                    }
                }
                for (Entry<Rule, Set<Integer>> entry : relevantRules.entrySet()) {
                    Rule rule = Desingularization.processRule(entry.getKey(), entry.getValue());
                    m_inputProcessor.processRule(rule);
                    m_writter.processRule(rule);
                }

                long durationQA = System.currentTimeMillis() - start;
                System.out.println(durationQA + " ms");
            }

        } catch (Exception exception) {
            exception.printStackTrace();
        }
        m_inputProcessor.endProcessing();
        m_writter.endProcessing();
    }

}
