package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;

import uk.ac.ox.cs.JRDFox.JRDFoxException;
import uk.ac.ox.cs.JRDFox.store.Resource;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.LabeledNull;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

public class RuleRelevanceUtilities {

    public static Map<Variable, Term> computeSubstitution(Atom groundAtom, Atom atom) {
        Map<Variable, Term> substitution = new LinkedHashMap<>();
        for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
            Map<Variable, Term> partialSubstitution = computeSubstitution(groundAtom.getArgument(argumentIndex),
                    atom.getArgument(argumentIndex));
            if (partialSubstitution != null && !verifyIfSubstitutionsConflict(substitution, partialSubstitution))
                substitution.putAll(partialSubstitution);
            else if (partialSubstitution == null || verifyIfSubstitutionsConflict(substitution, partialSubstitution))
                return null;
        }
        return substitution;
    }

    public static Map<Variable, Term> computeSubstitution(Term groundTerm, Term term) {
        Map<Variable, Term> substitution = new LinkedHashMap<>();
        if (groundTerm instanceof LabeledNull) {
            if (term instanceof Variable)
                substitution.put((Variable) term, groundTerm);
            else if (term instanceof LabeledNull && !groundTerm.equals(term) || term instanceof FunctionTerm
                    || term instanceof Constant)
                return null;
        } else if (groundTerm instanceof Constant) {
            if (term instanceof Variable)
                substitution.put((Variable) term, groundTerm);
            else if (term instanceof Constant && !groundTerm.equals(term) || term instanceof FunctionTerm
                    || term instanceof LabeledNull)
                return null;
        } else if (groundTerm instanceof FunctionTerm) {
            if (term instanceof Variable)
                substitution.put((Variable) term, groundTerm);
            else if (term instanceof Constant || term instanceof LabeledNull)
                return null;
            else if (term instanceof FunctionTerm
                    && ((FunctionTerm) groundTerm).getFunction().equals(((FunctionTerm) term).getFunction())) {
                for (int argumentIndex = 0; argumentIndex < ((FunctionTerm) groundTerm)
                        .getNumberOfArguments(); ++argumentIndex) {
                    Map<Variable, Term> partialSubstitution = computeSubstitution(
                            ((FunctionTerm) groundTerm).getArgument(argumentIndex),
                            ((FunctionTerm) term).getArgument(argumentIndex));
                    if (partialSubstitution != null
                            && !verifyIfSubstitutionsConflict(substitution, partialSubstitution))
                        substitution.putAll(partialSubstitution);
                    else if (partialSubstitution == null
                            || verifyIfSubstitutionsConflict(substitution, partialSubstitution))
                        return null;
                }
            } else
                return null;
        }
        return substitution;
    }

    public static boolean verifyIfSubstitutionsConflict(Map<Variable, Term> substitution1,
            Map<Variable, Term> substitution2) {
        for (Entry<Variable, Term> mapping : substitution1.entrySet()) {
            if (substitution2.containsKey(mapping.getKey())
                    && !substitution2.get(mapping.getKey()).equals(mapping.getValue()))
                return true;
        }
        return false;
    }

    public static Set<Map<Variable, Term>> computeSubstitutions(JRDFoxBridge jrdfoxBridge, List<Variable> answerVariables, List<Atom> queryAtoms, Function<Resource, ?> resourceValueFunction) throws JRDFoxException {
        Set<Map<Variable, Term>> substitutions = new LinkedHashSet<>();
        try (TupleIterator tupleIterator = jrdfoxBridge.createTupleIterator(false, answerVariables, queryAtoms, false)) {
            int arity = tupleIterator.getArity();
            for (long multiplicity = tupleIterator.open(); multiplicity != 0; multiplicity = tupleIterator.advance()) {
                Map<Variable, Term> bodySubstitution = new LinkedHashMap<>();
                for (int argumentIndex = 0; argumentIndex < arity; ++argumentIndex) {
                    Variable variable = answerVariables.get(argumentIndex);
                    bodySubstitution.put(variable, (Term) resourceValueFunction.apply(tupleIterator.getResource(argumentIndex)));
                }
                substitutions.add(bodySubstitution);
            }
        }
        return substitutions;
    }

    public static Atom[] applySubstitution(Atom[] atoms, Map<Variable, Term> substitution) {
        // Compute the substitutions for the body atoms
        Atom newBody[] = new Atom[atoms.length];
        for (int bodyAtomIndex = 0; bodyAtomIndex < atoms.length; ++bodyAtomIndex)
            newBody[bodyAtomIndex] = atoms[bodyAtomIndex].applySubstitution(substitution);
        return newBody;
    }

    public static boolean isTrivialEquality(Atom groundAtom) {
        if (Predicate.EQUALS.equals(groundAtom.getPredicate())
                && groundAtom.getArgument(0).equals(CommonUtilities.critical)
                && groundAtom.getArgument(1).equals(CommonUtilities.critical))
            return true;
        return false;
    }

    public static Rule replaceBodyConstantsWithCriticalConstant(Rule rule) {
        Atom[] newBodyAtoms = new Atom[rule.getNumberOfBodyAtoms()];
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
            Set<Constant> constants = new LinkedHashSet<>();
            bodyAtom.getConstants(constants);
            Map<Term, Term> mapping = new LinkedHashMap<>();
            for (Constant c : constants)
                mapping.put(c, CommonUtilities.critical);
            newBodyAtoms[bodyAtomIndex] = bodyAtom.applyMapping(mapping);
        }
        Rule newRule = Rule.create(rule.getHeadAtoms(), newBodyAtoms);
        return newRule;
    }

    public static Set<Rule> replaceBodyConstantsWithCriticalConstants(Set<Rule> rules) {
        Set<Rule> newRules = new LinkedHashSet<>();
        for (Rule rule : rules)
            newRules.add(replaceBodyConstantsWithCriticalConstant(rule));
        return newRules;
    }

    public static Rule createReplacementRuleForQueryLegacy(Atom query, Predicate goalPredicate) {
        Term[] newArguments = new Term[query.getNumberOfArguments()];
        Atom[] newBody = new Atom[query.getNumberOfArguments() + 1];
        int variableIndex = 0;
        int argumentIndex;
        for (argumentIndex = 0; argumentIndex < query.getNumberOfArguments(); ++argumentIndex) {
            Term argument = query.getArgument(argumentIndex);
            if (argument instanceof Constant || argument instanceof FunctionTerm)
                throw new java.lang.RuntimeException("Illegal argument for goal atom " + query);
            Variable newVariable = Variable.create("rr" + variableIndex++);
            newArguments[argumentIndex] = newVariable;
            newBody[argumentIndex] = Atom.create(Predicate.EQUALS, new Term[] { argument, newVariable });
        }
        newBody[argumentIndex] = Atom.create(query.getPredicate(), newArguments);
        return Rule.create(Atom.create(goalPredicate, query.getArguments()), newBody);
    }

    public static Map<Predicate, List<String>> createCriticalInstance(Map<Predicate, PredicateInfo> predicateInfo) {
        // Create the critical instance
        Map<Predicate, List<String>> criticalInstance = new LinkedHashMap<>();
        for (Predicate predicate : predicateInfo.keySet()) {
            PredicateInfo predicateSchema = predicateInfo.get(predicate);
            if (!predicateSchema.isIDB()) {
                List<String> argumentLexicalForms = new ArrayList<>();
                for (int column = 0; column < predicateSchema.getArity(); ++column)
                    argumentLexicalForms.add(CommonUtilities.criticalLexicalForm);
                criticalInstance.put(predicate, argumentLexicalForms);
            }
        }
        return criticalInstance;
    }

    public static Set<Map<Variable, Term>> computeAllValidSubstitutions(JRDFoxBridge jrdfoxBridge, Rule rule, Atom fact) throws JRDFoxException {
        Set<Map<Variable, Term>> substitutions = new LinkedHashSet<>();
        // Compute the substitution for the head atom
        Map<Variable, Term> headSubstitution = RuleRelevanceUtilities.computeSubstitution(fact, rule.getHeadAtom(0));
        if (headSubstitution != null) {
            // Apply the substitutions to the body atoms
            Atom[] newBody = RuleRelevanceUtilities.applySubstitution(rule.getBodyAtoms(), headSubstitution);

            // Compute the substitutions for the body atoms
            Set<Variable> newHeadArguments = new LinkedHashSet<>();
            for (int bodyAtomIndex = 0; bodyAtomIndex < newBody.length; ++bodyAtomIndex)
                newBody[bodyAtomIndex].getFreeVariables(newHeadArguments);

            if (newHeadArguments.isEmpty()) {
                // Check if the body atoms are in the model
                Boolean inModel = true;
                for (Atom bodyAtom : newBody) {
                    int arity = bodyAtom.getNumberOfArguments();
                    Variable[] arguments = new Variable[arity];
                    for (int column = 0; column < arity; ++column)
                        arguments[column] = Variable.create("X" + column);
                    Atom atom = Atom.create(bodyAtom.getPredicate(), arguments);

                    Map<Variable, Term> goldSubstitution = new LinkedHashMap<>();
                    for (int column = 0; column < arity; ++column)
                        goldSubstitution.put(Variable.create("X" + column), bodyAtom.getArgument(column));

                    Set<Map<Variable, Term>> bodySubstitutions = RuleRelevanceUtilities.computeSubstitutions(jrdfoxBridge, Arrays.asList(arguments), Arrays.asList(atom),
                        (x) -> {
                            return jrdfoxBridge.parseResource(x);
                        }
                    );
                    if (!bodySubstitutions.contains(goldSubstitution)) {
                        inModel = false;
                        break;
                    }
                }
                if (inModel)
                    substitutions.add(headSubstitution);
            } else {
                substitutions = RuleRelevanceUtilities.computeSubstitutions(jrdfoxBridge, new ArrayList<>(newHeadArguments), Arrays.asList(newBody),
                    (x) -> {
                        return jrdfoxBridge.parseResource(x);
                    }
                );
                for (Map<Variable, Term> s : substitutions) {
                    s.putAll(headSubstitution);
                }
            }
        }
        return substitutions;
    }
}
