package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class RuleVerifier implements InputProcessor {
    protected final InputProcessor m_inputProcessor;

    public RuleVerifier(InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
    }

    public void processRule(Rule rule) {
        assertRuleHasNoCartesianProducts(rule);
        assertAllEqualtiesAreGuarded(rule);
        assertEachHeadVariableOccursInBody(rule);
        m_inputProcessor.processRule(rule);
    }

    protected void assertRuleHasNoCartesianProducts(Rule rule) {
        Set<Variable> variables = new LinkedHashSet<>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            if (bodyAtomIndex > 0) {
                Set<Variable> newVariables = new LinkedHashSet<>();
                rule.getBodyAtom(bodyAtomIndex).getFreeVariables(newVariables);
                if (variables.size() > 0 && Collections.disjoint(variables, newVariables))
                    throw new RuntimeException("Rule " + rule + " has Cartesian products");
            }
            rule.getBodyAtom(bodyAtomIndex).getFreeVariables(variables);
        }
    }

    protected void assertAllEqualtiesAreGuarded(Rule rule) {
        Set<Variable> variables = new LinkedHashSet<>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom atom = rule.getBodyAtom(bodyAtomIndex);
            if (!atom.getPredicate().equals(Predicate.EQUALS))
                atom.getFreeVariables(variables);
        }

        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom atom = rule.getBodyAtom(bodyAtomIndex);
            if (atom.getPredicate().equals(Predicate.EQUALS)) {
                Set<Variable> newVariables = new LinkedHashSet<>();
                atom.getFreeVariables(newVariables);
                if (!variables.containsAll(newVariables))
                    throw new RuntimeException("Equality " + rule + " is not guarded");
            }
        }
    }

    protected void assertEachHeadVariableOccursInBody(Rule rule) {
        Set<Variable> bodyVariables = new LinkedHashSet<>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex)
            rule.getBodyAtom(bodyAtomIndex).getFreeVariables(bodyVariables);
        Set<Variable> headVariables = new LinkedHashSet<>();
        rule.getHeadAtom(0).getFreeVariables(headVariables);
        if (!bodyVariables.containsAll(headVariables))
            throw new RuntimeException("Rule " + rule + " is unfound");
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        m_inputProcessor.endProcessing();
    }

}
