package uk.ac.ox.cs.goalDriven.QA.utilities;

import java.util.AbstractMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Arrays;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import uk.ac.ox.cs.JRDFox.JRDFoxException;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridgeUtilities;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

import java.util.Set;

public class CommonUtilities {

    protected static Variable X = Variable.create("X");
    protected static Variable Y = Variable.create("Y");
    protected static Variable Z = Variable.create("Z");
    public static Rule transitivity = Rule.create(
            Atom.create(Predicate.EQUALS, new Variable[] { X, Z }),
            new Atom[] {
                    Atom.create(Predicate.EQUALS,
                            new Variable[] { X, Y }),
                    Atom.create(Predicate.EQUALS,
                            new Variable[] { Y, Z }) });
    public static Rule symmetry = Rule.create(
            Atom.create(Predicate.EQUALS, new Variable[] { X, Y }),
            new Atom[] {
                    Atom.create(Predicate.EQUALS,
                            new Variable[] { Y, X }) });

    public static String criticalLexicalForm = "c";
    public static Constant critical = Constant.create(criticalLexicalForm);

    public static Set<Rule> createReflexivityAxioms(Map<Predicate, PredicateInfo> predicateInfo) {
        Set<Rule> reflexivityAxioms = new LinkedHashSet<>();
        for (Predicate predicate : predicateInfo.keySet()) {
            if (!predicate.equals(Predicate.EQUALS)) {
                PredicateInfo databaseSchema = predicateInfo.get(predicate);
                Variable[] arguments = new Variable[databaseSchema.getArity()];
                for (int column = 0; column < databaseSchema.getArity(); ++column)
                    arguments[column] = Variable.create("X" + column);

                Atom body = Atom.create(predicate, arguments);
                for (int column = 0; column < databaseSchema.getArity(); ++column) {
                    Rule reflexivity = Rule.create(Atom.create(Predicate.EQUALS,
                            new Variable[] { arguments[column], arguments[column] }), body);
                    reflexivityAxioms.add(reflexivity);
                }
            }
        }
        return reflexivityAxioms;
    }

    public static class PredicateInfo {
        protected final boolean m_isIDB;
        protected final int m_arity;

        public PredicateInfo(boolean isIDB, int arity) {
            m_isIDB = isIDB;
            m_arity = arity;
        }

        public int getArity() {
            return m_arity;
        }

        public boolean isIDB() {
            return m_isIDB;
        }
    }

    public static Map<Predicate, PredicateInfo> getPredicateInfo(Set<Rule> rules) {
        Map<Predicate, Entry<Boolean, Integer>> predicateInfo = new LinkedHashMap<>();
        for (Rule rule : rules) {
            for (int headAtomIndex = 0; headAtomIndex < rule.getNumberOfHeadAtoms(); ++headAtomIndex) {
                Predicate predicate = rule.getHeadAtom(headAtomIndex).getPredicate();
                int arity = rule.getHeadAtom(headAtomIndex).getNumberOfArguments();
                boolean isIDB = true;
                if (predicateInfo.containsKey(predicate) && predicateInfo.get(predicate).getValue() != arity)
                    throw new RuntimeException(
                            "Predicate " + predicate.getName() + " occurs with arity " + Integer.toString(arity));
                predicateInfo.put(predicate, new AbstractMap.SimpleEntry<Boolean, Integer>(isIDB, arity));
            }

            for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
                Predicate predicate = rule.getBodyAtom(bodyAtomIndex).getPredicate();
                int arity = rule.getBodyAtom(bodyAtomIndex).getNumberOfArguments();
                boolean isIDB = false;
                if (predicateInfo.containsKey(predicate) && predicateInfo.get(predicate).getValue() != arity)
                    throw new RuntimeException(
                            "Predicate " + predicate.getName() + " occurs with arity " + Integer.toString(arity));
                if (!predicateInfo.containsKey(predicate))
                    predicateInfo.put(predicate, new AbstractMap.SimpleEntry<Boolean, Integer>(isIDB, arity));
            }
        }

        Map<Predicate, PredicateInfo> ret = new LinkedHashMap<>();
        for (Map.Entry<Predicate, Entry<Boolean, Integer>> entry : predicateInfo.entrySet()) {
            Predicate predicate = entry.getKey();
            Entry<Boolean, Integer> info = entry.getValue();
            ret.put(predicate, new PredicateInfo(info.getKey(), info.getValue()));
        }
        return ret;
    }

    public static Map<Predicate, Set<Rule>> createRuleIndex(Set<Rule> rules) {
        // Index the rules based on their head atoms
        Map<Predicate, Set<Rule>> indexedRules = new LinkedHashMap<>();
        for (Rule rule : rules) {
            if (rule.getHeadAtoms().length > 1)
                throw new java.lang.RuntimeException("Non datalog rules are not supported");
            else {
                Set<Rule> indexed = indexedRules.get(rule.getHeadAtom(0).getPredicate());
                if (indexed == null) {
                    indexedRules.put(rule.getHeadAtom(0).getPredicate(), new LinkedHashSet<>());
                    indexedRules.get(rule.getHeadAtom(0).getPredicate()).add(rule);
                } else
                    indexed.add(rule);
            }
        }
        return indexedRules;
    }

    public static Integer[] saveInstanceOnDisk(JRDFoxBridge jrdfoxBridge, String path,
            Set<Rule> rules, List<Predicate> predicatesToSave) throws JRDFoxException, IOException {
        int total_facts = 0;
        int total_auxiliary_facts = 0;
        int total_magic_facts = 0;
        new File(path).mkdirs();
        Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(rules);
        for (Entry<Predicate, PredicateInfo> entry : predicateInfo.entrySet()) {
            Predicate predicate = entry.getKey();

            PredicateInfo info = entry.getValue();
            int arity = info.getArity();
            Variable[] terms = new Variable[arity];
            for (int i = 0; i < arity; ++i)
                terms[i] = Variable.create("X" + i);
            Atom atom = Atom.create(predicate, terms);

            List<List<String>> answers = JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(terms),
                    Arrays.asList(atom),
                    (x) -> {
                        return jrdfoxBridge.parseResource(x);
                    }, false, false);

            total_facts += answers.size();
            if (predicate.getName().startsWith("F_")
                    || predicate.equals(D))
                total_auxiliary_facts += answers.size();

            if (predicate.getName().startsWith(uk.ac.ox.cs.goalDriven.QA.magic.MagicUtilities.magicPrefix))
                // if (predicate.getName().startsWith("mgc_"))
                total_magic_facts += answers.size();

            if (predicatesToSave == null || predicatesToSave.contains(predicate)) {
                FileWriter writer = new FileWriter(path + "/" + predicate.getName() + ".csv");
                for (int row = 0; row < answers.size(); ++row) {
                    for (int column = 0; column < answers.get(row).size(); ++column) {
                        writer.write(answers.get(row).get(column));
                        if (column < answers.get(row).size() - 1)
                            writer.write(",");
                    }
                    writer.write("\n");
                }
                writer.close();
            }
        }
        return new Integer[] { total_facts, total_magic_facts, total_auxiliary_facts };
    }

    public static Predicate D = Predicate.create("D");

    // D(X) -> X=X
    private static Atom DX = Atom.create(D, X);
    private static Atom EQXX = Atom.create(Predicate.EQUALS, X, X);
    public static Rule R = Rule.create(EQXX, DX);

    public static Set<Rule> getDAxiomatization(Set<Rule> program) {

        Set<Rule> axioms = new LinkedHashSet<>();

        // Get all constants
        Set<Constant> constants = new LinkedHashSet<>();
        // Get all function symbols
        for (Rule rule : program) {
            for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
                Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
                for (int i = 0; i < bodyAtom.getNumberOfArguments(); ++i)
                    bodyAtom.getArgument(i).getConstants(constants);

            }

            for (int headAtomIndex = 0; headAtomIndex < rule.getNumberOfHeadAtoms(); ++headAtomIndex) {
                Atom headAtom = rule.getHeadAtom(headAtomIndex);
                for (int i = 0; i < headAtom.getNumberOfArguments(); ++i)
                    headAtom.getArgument(i).getConstants(constants);

            }
        }

        // The below is commented as we already add the axioms R(...,X,...) -> D(X)
        // D(P) = {D(c) | c occurs in P}
        // for (Constant c : constants)
        // axioms.add(Rule.create(Atom.create(D, c)));

        // R(X1, \dots, Xk, \dots, Xn) -> D(Xk)
        axioms.addAll(getDomainAxioms(program));
        return axioms;

    }

    // R(X1, \dots, Xk, \dots, Xn) -> D(Xk)
    public static Set<Rule> getDomainAxioms(Set<Rule> program) {
        Set<Rule> axioms = new LinkedHashSet<>();
        Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(program);
        for (Map.Entry<Predicate, PredicateInfo> entry : predicateInfo.entrySet()) {
            Predicate predicate = entry.getKey();
            PredicateInfo info = entry.getValue();
            // @WARNING Query names should be passed as an argument in the input
            if (!predicate.equals(Predicate.EQUALS) && !predicate.getName().startsWith("Q")
                    && !(predicate.getName().startsWith("q") && predicate.getName().charAt(1) >= '0'
                            && predicate.getName().charAt(1) <= '9')
                    && !predicate.getName().startsWith("F_")
                    && !predicate.getName().startsWith(uk.ac.ox.cs.goalDriven.QA.magic.MagicUtilities.magicPrefix)
                    && !predicate.equals(D)) {
                Term[] terms = new Term[info.getArity()];
                for (int i = 0; i < info.getArity(); ++i)
                    terms[i] = Variable.create("X" + Integer.toString(i));
                Atom bodyAtom = Atom.create(predicate, terms);
                for (int i = 0; i < info.getArity(); ++i) {
                    Atom DX = Atom.create(D, Variable.create("X" + Integer.toString(i)));
                    axioms.add(Rule.create(DX, bodyAtom));
                }
            }
        }
        return axioms;
    }

    public static Set<Rule> getFRAxiomatization(Map<Function, Integer> trueFunctions) {
        Set<Rule> axioms = new LinkedHashSet<>();

        // \bigwedge_k D(X_k), X_k = Y_k, D(Y_k) -> f(X) = f(Y)
        for (Map.Entry<Function, Integer> entry : trueFunctions.entrySet()) {
            Function function = entry.getKey();
            Integer arity = entry.getValue();
            Term[] Xterms = new Term[arity];
            Term[] Yterms = new Term[arity];
            for (int i = 0; i < arity; ++i) {
                Xterms[i] = Variable.create("X" + Integer.toString(i));
                Yterms[i] = Variable.create("Y" + Integer.toString(i));
            }
            Term fX1Xk = FunctionTerm.create(function, Xterms);
            Term fY1Yk = FunctionTerm.create(function, Yterms);
            Atom EQfXfY = Atom.create(Predicate.EQUALS, fX1Xk, fY1Yk);
            Atom[] bodyAtoms = new Atom[3 * arity];
            for (int i = 0; i < arity; ++i) {
                bodyAtoms[3 * i] = Atom.create(D, Xterms[i]);
                bodyAtoms[3 * i + 1] = Atom.create(Predicate.EQUALS, Xterms[i], Yterms[i]);
                bodyAtoms[3 * i + 2] = Atom.create(D, Yterms[i]);
            }
            axioms.add(Rule.create(EQfXfY, bodyAtoms));
        }

        return axioms;
    }

    public static Map<Function, Integer> getTrueFunctions(Set<Rule> rules) {
        Map<Function, Integer> trueFunctions = new LinkedHashMap<>();
        for (Rule rule : rules) {
            for (int atomIndex = 0; atomIndex < rule.getNumberOfHeadAtoms(); ++atomIndex) {
                Atom atom = rule.getHeadAtom(atomIndex);
                for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
                    Term term = atom.getArgument(argumentIndex);
                    if (term instanceof FunctionTerm)
                        trueFunctions.put(((FunctionTerm) term).getFunction(),
                                ((FunctionTerm) term).getNumberOfArguments());
                }
            }
            for (int atomIndex = 0; atomIndex < rule.getNumberOfBodyAtoms(); ++atomIndex) {
                Atom atom = rule.getBodyAtom(atomIndex);
                for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
                    Term term = atom.getArgument(argumentIndex);
                    if (term instanceof FunctionTerm)
                        trueFunctions.put(((FunctionTerm) term).getFunction(),
                                ((FunctionTerm) term).getNumberOfArguments());
                }
            }
        }
        return trueFunctions;
    }

}
