package uk.ac.ox.cs.goalDriven.RDFoxBridge;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.JRDFox.JRDFoxException;
import uk.ac.ox.cs.JRDFox.SimplePrefixes;
import uk.ac.ox.cs.JRDFox.model.Bind;
import uk.ac.ox.cs.JRDFox.model.BuiltinFunctionCall;
import uk.ac.ox.cs.JRDFox.model.Datatype;
import uk.ac.ox.cs.JRDFox.model.Individual;
import uk.ac.ox.cs.JRDFox.model.Literal;
import uk.ac.ox.cs.JRDFox.model.LiteralValue;
import uk.ac.ox.cs.JRDFox.store.DataStore;
import uk.ac.ox.cs.JRDFox.store.Dictionary;
import uk.ac.ox.cs.JRDFox.store.Resource;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.LabeledNull;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class JRDFoxBridge implements AutoCloseable {       
    protected final DataStore m_dataStore;
    protected final Dictionary m_dictionary;
    
    public JRDFoxBridge(boolean optimizeEquality) throws JRDFoxException {
        Map<String, String> parameters = new HashMap<String, String>();
        if (optimizeEquality)
            parameters.put("equality", "noUNAnoRef");
        m_dataStore = new DataStore("par-complex-nn", parameters);
        m_dictionary = m_dataStore.getDictionary();
    }

    public void close() {
        m_dataStore.dispose();
    }

    public DataStore getDataStore() {
    	return m_dataStore;
    }
    
    static protected uk.ac.ox.cs.JRDFox.model.Term convertTerm(Term term, Collection<Literal> body, Map<FunctionTerm, uk.ac.ox.cs.JRDFox.model.Variable> variablesByFunctionTerms) {
        if (term instanceof Variable)
            return uk.ac.ox.cs.JRDFox.model.Variable.create(((Variable)term).getName());
        else if (term instanceof Constant)
            return uk.ac.ox.cs.JRDFox.model.LiteralValue.create(((Constant)term).getLexicalForm(), Datatype.XSD_STRING);
        else if (term instanceof LabeledNull)
            return uk.ac.ox.cs.JRDFox.model.BlankNode.create("__ln__" + ((LabeledNull)term).getName());
        else if (term instanceof FunctionTerm) {
            FunctionTerm functionTerm = (FunctionTerm)term;
            var functionTermVariable = variablesByFunctionTerms.get(functionTerm);
            if (functionTermVariable == null) {
                functionTermVariable = uk.ac.ox.cs.JRDFox.model.Variable.create("__fnterm__" + variablesByFunctionTerms.size()); 
                variablesByFunctionTerms.put(functionTerm, functionTermVariable);
                int numberOfArguments = functionTerm.getNumberOfArguments();
                var skolemArguments = new uk.ac.ox.cs.JRDFox.model.Term[numberOfArguments + 1];
                skolemArguments[0] = LiteralValue.create(functionTerm.getFunction().getName(), Datatype.XSD_STRING);
                for (int index = 0; index < numberOfArguments; ++index)
                    skolemArguments[index + 1] = convertTerm(functionTerm.getArgument(index), body, variablesByFunctionTerms);
                body.add(Bind.create(BuiltinFunctionCall.create("SKOLEM", skolemArguments), functionTermVariable));
            }
            return functionTermVariable;
        }
        else
            throw new RuntimeException("Unexpected argument form.");
    }

    static protected Individual toNullaryIndividual(Predicate predicate) {
        return Individual.create("nullary:" + predicate.getName());
    }
    
    static protected Individual toUnaryIndividual(Predicate predicate) {
        return Individual.create("unary:" + predicate.getName());
    }
    
    static protected Individual toBinaryIndividual(Predicate predicate) {
    	if (Predicate.EQUALS.equals(predicate))
    		return Individual.SAME_AS;
    	else
    		return Individual.create("binary:" + predicate.getName());
    }

    static protected Individual toNAryIndividual(Predicate predicate, int position) {
        return Individual.create("nary:" + predicate.getName() + ":" + position);
    }

    static protected void convertBodyAtom(Atom atom, Collection<Literal> body, Map<FunctionTerm, uk.ac.ox.cs.JRDFox.model.Variable> variablesByFunctionTerms) {
        int numberOfArguments = atom.getNumberOfArguments();
        if (numberOfArguments == 0)
            body.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(ABOUT_NULLARY, Individual.RDF_TYPE, toNullaryIndividual(atom.getPredicate())));
        else if (numberOfArguments == 1)
            body.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(convertTerm(atom.getArgument(0), body, variablesByFunctionTerms), Individual.RDF_TYPE, toUnaryIndividual(atom.getPredicate())));
        else if (numberOfArguments == 2)
            body.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(convertTerm(atom.getArgument(0), body, variablesByFunctionTerms), toBinaryIndividual(atom.getPredicate()), convertTerm(atom.getArgument(1), body, variablesByFunctionTerms)));
        else {
            var tupleVariable = uk.ac.ox.cs.JRDFox.model.Variable.create("__body__" + body.size()); 
            for (int index = 0; index < numberOfArguments; ++index) {
                var argument = convertTerm(atom.getArgument(index), body, variablesByFunctionTerms);
                body.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(tupleVariable, toNAryIndividual(atom.getPredicate(), index), argument));
            }
        }
    }
    
    static protected void convertHeadAtom(Atom atom, Collection<uk.ac.ox.cs.JRDFox.model.Atom> head, Collection<Literal> body, Map<FunctionTerm, uk.ac.ox.cs.JRDFox.model.Variable> variablesByFunctionTerms) {
        int numberOfArguments = atom.getNumberOfArguments();
        if (numberOfArguments == 0)
            head.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(ABOUT_NULLARY, Individual.RDF_TYPE, toNullaryIndividual(atom.getPredicate())));
        else if (numberOfArguments == 1)
        	head.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(convertTerm(atom.getArgument(0), body, variablesByFunctionTerms), Individual.RDF_TYPE, toUnaryIndividual(atom.getPredicate())));
        else if (numberOfArguments == 2)
        	head.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(convertTerm(atom.getArgument(0), body, variablesByFunctionTerms), toBinaryIndividual(atom.getPredicate()), convertTerm(atom.getArgument(1), body, variablesByFunctionTerms)));
        else {
            var skolemArguments = new uk.ac.ox.cs.JRDFox.model.Term[numberOfArguments + 1];
            skolemArguments[0] = LiteralValue.create(atom.getPredicate().getName(), Datatype.XSD_STRING);
            var tupleVariable = uk.ac.ox.cs.JRDFox.model.Variable.create("__head__" + head.size()); 
            for (int index = 0; index < numberOfArguments; ++index) {
                var argument = convertTerm(atom.getArgument(index), body, variablesByFunctionTerms);
                head.add(uk.ac.ox.cs.JRDFox.model.Atom.rdf(tupleVariable, toNAryIndividual(atom.getPredicate(), index), argument));
                skolemArguments[index + 1] = argument;
            }
            body.add(Bind.create(BuiltinFunctionCall.create("SKOLEM", skolemArguments), tupleVariable));
        }
    }

    static public uk.ac.ox.cs.JRDFox.model.Rule convertRule(Rule rule) {
        var body = new ArrayList<Literal>();
        var head = new ArrayList<uk.ac.ox.cs.JRDFox.model.Atom>();
        var variablesByFunctionTerms = new HashMap<FunctionTerm, uk.ac.ox.cs.JRDFox.model.Variable>();
        for (Atom bodyAtom : rule.getBodyAtoms())
            convertBodyAtom(bodyAtom, body, variablesByFunctionTerms);
        for (Atom headAtom : rule.getHeadAtoms())
            convertHeadAtom(headAtom, head, body, variablesByFunctionTerms);
        return uk.ac.ox.cs.JRDFox.model.Rule.create(head, body);
    }

    public void addRules(Collection<Rule> rules) throws JRDFoxException {
        var convertedRules = new ArrayList<uk.ac.ox.cs.JRDFox.model.Rule>();
        for (Rule rule : rules)
            convertedRules.add(convertRule(rule));
        m_dataStore.importRules(convertedRules);
    }
    
    public void computeMaterialization() throws JRDFoxException {
        m_dataStore.updateStatistics();
        m_dataStore.applyReasoning(true);
    }
    
    static public String convertQuery(boolean distinct, List<Variable> answerVariables, List<Atom> queryAtoms, boolean onlyCertainAnswers) throws JRDFoxException {
        var convertedQueryAtoms = new ArrayList<Literal>();
        var variablesByFunctionTerms = new HashMap<FunctionTerm, uk.ac.ox.cs.JRDFox.model.Variable>();
        for (Atom atom : queryAtoms)
            convertBodyAtom(atom, convertedQueryAtoms, variablesByFunctionTerms);
        StringBuilder queryText = new StringBuilder();
        queryText.append("SELECT ");
        if (distinct)
            queryText.append("DISTINCT ");
        for (Variable variable : answerVariables) {
            queryText.append('?');
            queryText.append(variable.getName());
            queryText.append(' ');
        }
        queryText.append("WHERE { ");
        for (Literal literal : convertedQueryAtoms) {
            if (literal instanceof uk.ac.ox.cs.JRDFox.model.Atom) {
                literal.getArgument(0).toString(queryText, SimplePrefixes.EMPTY_IMMUTABLE_INSTANCE);
                queryText.append(' ');
                literal.getArgument(1).toString(queryText, SimplePrefixes.EMPTY_IMMUTABLE_INSTANCE);
                queryText.append(' ');
                literal.getArgument(2).toString(queryText, SimplePrefixes.EMPTY_IMMUTABLE_INSTANCE);
            }
            else
                literal.toString(queryText, SimplePrefixes.EMPTY_IMMUTABLE_INSTANCE);
            queryText.append(" . ");
        }
        if (onlyCertainAnswers) {
            for (Variable variable : answerVariables) {
                queryText.append("FILTER(isLiteral(?");
                queryText.append(variable.getName());
                queryText.append(")) . ");
            }
        }
        queryText.append('}');
        return queryText.toString();
    }
    
    public TupleIterator createTupleIterator(boolean distinct, List<Variable> answerVariables, List<Atom> queryAtoms, boolean onlyCertainAnswers) throws JRDFoxException {
        String queryText = convertQuery(distinct, answerVariables, queryAtoms, onlyCertainAnswers);
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("domain", "IDBrep");
        return m_dataStore.compileQuery(queryText, SimplePrefixes.EMPTY_IMMUTABLE_INSTANCE, parameters);
    }
    
    protected Term parseFunctionEncoding(String string) {
    	int length = string.length();
    	if (length == 0)
    		throw new RuntimeException("The exmpty string does not correspond to a function encoding.");
    	int index = 0;
    	while (index < length && string.charAt(index) != '_')
    		++index;
    	Function function = Function.create(string.substring(0, index));
    	List<Term> arguments = new ArrayList<Term>();
    	while (index < length) {
    		++index;
    		int start = index;
        	while (index < length && string.charAt(index) != '_')
        		++index;
        	String resourceIDText = string.substring(start, index);
        	long resourceID = Long.parseLong(resourceIDText);
        	try {
        		arguments.add(parseResource(m_dictionary.getResource(resourceID)));
        	}
        	catch (JRDFoxException e) {
        		throw new RuntimeException("An error occurred while resolving resource ID " + resourceID + " in the dictionary.", e);
        	}
    	}
        return FunctionTerm.create(function, arguments.toArray(new Term[arguments.size()]));
    }
    
    public Term parseResource(Resource resource) {
    	switch (resource.m_datatype) {
    	case XSD_STRING:
    		return Constant.create(resource.m_lexicalForm);
    	case BLANK_NODE:
    		if (resource.m_lexicalForm.startsWith("__ln__"))
    			return LabeledNull.create(resource.m_lexicalForm.substring(6));
    		else
    			return parseFunctionEncoding(resource.m_lexicalForm);
    	default:
    		throw new RuntimeException("Unexpected datatype in the answer.");
    	}
    }

    public InputProcessor newInputProcessor() {
        return new JRDFoxInputProcessor(m_dataStore);
    }

    protected static class JRDFoxInputProcessor implements InputProcessor {
        protected static final String[] U_ESCAPE = {
            "\\u0000", "\\u0001", "\\u0002", "\\u0003", "\\u0004", "\\u0005", "\\u0006", "\\u0007", "\\b",     "\\t",     "\\n",     "\\u000B", "\\f",     "\\r",     "\\u000E", "\\u000F",
            "\\u0010", "\\u0011", "\\u0012", "\\u0013", "\\u0014", "\\u0015", "\\u0016", "\\u0017", "\\u0018", "\\u0019", "\\u001A", "\\u001B", "\\u001C", "\\u001D", "\\u001E", "\\u001F",
            "\\u0020", "\\u0021", "\\u0022", "\\u0023", "\\u0024", "\\u0025", "\\u0026", "\\u0027", "\\u0028", "\\u0029", "\\u002A", "\\u002B", "\\u002C", "\\u002D", "\\u002E", "\\u002F",
            "\\u0030", "\\u0031", "\\u0032", "\\u0033", "\\u0034", "\\u0035", "\\u0036", "\\u0037", "\\u0038", "\\u0039", "\\u003A", "\\u003B", "\\u003C", "\\u003D", "\\u003E", "\\u003F",
            "\\u0040", "\\u0041", "\\u0042", "\\u0043", "\\u0044", "\\u0045", "\\u0046", "\\u0047", "\\u0048", "\\u0049", "\\u004A", "\\u004B", "\\u004C", "\\u004D", "\\u004E", "\\u004F",
            "\\u0050", "\\u0051", "\\u0052", "\\u0053", "\\u0054", "\\u0055", "\\u0056", "\\u0057", "\\u0058", "\\u0059", "\\u005A", "\\u005B", "\\u005C", "\\u005D", "\\u005E", "\\u005F",
            "\\u0060", "\\u0061", "\\u0062", "\\u0063", "\\u0064", "\\u0065", "\\u0066", "\\u0067", "\\u0068", "\\u0069", "\\u006A", "\\u006B", "\\u006C", "\\u006D", "\\u006E", "\\u006F",
            "\\u0070", "\\u0071", "\\u0072", "\\u0073", "\\u0074", "\\u0075", "\\u0076", "\\u0077", "\\u0078", "\\u0079", "\\u007A", "\\u007B", "\\u007C", "\\u007D", "\\u007E", "\\u007F",
            "\\u0080", "\\u0081", "\\u0082", "\\u0083", "\\u0084", "\\u0085", "\\u0086", "\\u0087", "\\u0088", "\\u0089", "\\u008A", "\\u008B", "\\u008C", "\\u008D", "\\u008E", "\\u008F",
            "\\u0090", "\\u0091", "\\u0092", "\\u0093", "\\u0094", "\\u0095", "\\u0096", "\\u0097", "\\u0098", "\\u0099", "\\u009A", "\\u009B", "\\u009C", "\\u009D", "\\u009E", "\\u009F",
            "\\u00A0", "\\u00A1", "\\u00A2", "\\u00A3", "\\u00A4", "\\u00A5", "\\u00A6", "\\u00A7", "\\u00A8", "\\u00A9", "\\u00AA", "\\u00AB", "\\u00AC", "\\u00AD", "\\u00AE", "\\u00AF",
            "\\u00B0", "\\u00B1", "\\u00B2", "\\u00B3", "\\u00B4", "\\u00B5", "\\u00B6", "\\u00B7", "\\u00B8", "\\u00B9", "\\u00BA", "\\u00BB", "\\u00BC", "\\u00BD", "\\u00BE", "\\u00BF",
            "\\u00C0", "\\u00C1", "\\u00C2", "\\u00C3", "\\u00C4", "\\u00C5", "\\u00C6", "\\u00C7", "\\u00C8", "\\u00C9", "\\u00CA", "\\u00CB", "\\u00CC", "\\u00CD", "\\u00CE", "\\u00CF",
            "\\u00D0", "\\u00D1", "\\u00D2", "\\u00D3", "\\u00D4", "\\u00D5", "\\u00D6", "\\u00D7", "\\u00D8", "\\u00D9", "\\u00DA", "\\u00DB", "\\u00DC", "\\u00DD", "\\u00DE", "\\u00DF",
            "\\u00E0", "\\u00E1", "\\u00E2", "\\u00E3", "\\u00E4", "\\u00E5", "\\u00E6", "\\u00E7", "\\u00E8", "\\u00E9", "\\u00EA", "\\u00EB", "\\u00EC", "\\u00ED", "\\u00EE", "\\u00EF",
            "\\u00F0", "\\u00F1", "\\u00F2", "\\u00F3", "\\u00F4", "\\u00F5", "\\u00F6", "\\u00F7", "\\u00F8", "\\u00F9", "\\u00FA", "\\u00FB", "\\u00FC", "\\u00FD", "\\u00FE", "\\u00FF"
        };
        
        protected static final Datatype[] XSD_STRING = new Datatype[] { Datatype.XSD_STRING };
        
        protected final DataStore m_dataStore;
        protected final Dictionary m_dictionary;
        protected final List<uk.ac.ox.cs.JRDFox.model.Rule> m_rules;
        protected final StringBuilder m_facts;
        protected final StringBuilder m_skolemID;
        protected final StringBuilder m_temporary;
        protected final String[] m_lexicalForm;
        protected int m_numberOfQueuedFacts;
        
        public JRDFoxInputProcessor(DataStore dataStore) {
            m_dataStore = dataStore;
            m_dictionary = m_dataStore.getDictionary();
            m_rules = new ArrayList<uk.ac.ox.cs.JRDFox.model.Rule>();
            m_facts = new StringBuilder();
            m_skolemID = new StringBuilder();
            m_temporary = new StringBuilder();
            m_lexicalForm = new String[1];
            m_numberOfQueuedFacts = 0;
        }
        
        public void startProcessing() {
            m_rules.clear();
            m_facts.delete(0, m_facts.length());
            m_numberOfQueuedFacts = 0;
        }
        
        public void processRule(Rule rule) {
            m_rules.add(convertRule(rule));
        }

        protected void appendFactLexicalForm(StringBuilder buffer, String string) {
            int stringLength = string.length();
            for (int index = 0; index < stringLength; ++index) {
                char c = string.charAt(index);
                if (c < 32)
                    buffer.append(U_ESCAPE[c]);
                else {
                    if (c == '"' || c == '\\')
                        buffer.append('\\');
                    buffer.append(c);
                }
            }
        }

        protected void appendFactString(StringBuilder buffer, String string) {
            buffer.append('"');
            appendFactLexicalForm(buffer, string);
            buffer.append('"');
        }
        
        public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
            try {
                switch (argumentLexicalForms.size()) {
                case 0:
                    m_facts.append("<about:nullary> a <nullary:");
                    m_facts.append(predicate.getName());
                    m_facts.append("> .\n");
                    break;
                case 1:
                    appendFactString(m_facts, argumentLexicalForms.get(0));
                    m_facts.append(" a <unary:");
                    m_facts.append(predicate.getName());
                    m_facts.append("> .\n");
                    break;
                case 2:
                    appendFactString(m_facts, argumentLexicalForms.get(0));
                    m_facts.append(" <binary:");
                    m_facts.append(predicate.getName());
                    m_facts.append("> ");
                    appendFactString(m_facts, argumentLexicalForms.get(1));
                    m_facts.append(" .\n");
                    break;
                default:
                    m_skolemID.delete(0, m_skolemID.length());
                    m_skolemID.append("_:");
                    m_skolemID.append(predicate.getName());
                    int arity = argumentLexicalForms.size();
                    for (int index = 0; index < arity; ++index) {
                        m_temporary.delete(0, m_temporary.length());
                        appendFactLexicalForm(m_temporary, argumentLexicalForms.get(index));
                        m_lexicalForm[0] = m_temporary.toString();
                        long[] resourceIDs = m_dictionary.resolveResources(m_lexicalForm, XSD_STRING);
                        m_skolemID.append('_');
                        m_skolemID.append(resourceIDs[0]);
                    }
                    for (int index = 0; index < arity; ++index) {
                        if (index > 0)
                            m_facts.append(' ');
                        m_facts.append(m_skolemID);
                        m_facts.append(" <nary:");
                        m_facts.append(predicate.getName());
                        m_facts.append(':');
                        m_facts.append(index);
                        m_facts.append("> ");
                        appendFactString(m_facts, argumentLexicalForms.get(index));
                        m_facts.append(" .");
                    }
                    m_facts.append('\n');
                    break;
                }
                ++m_numberOfQueuedFacts;
                if (m_numberOfQueuedFacts == 1000) {
                    m_dataStore.importText(m_facts.toString());
                    m_facts.delete(0, m_facts.length());
                    m_numberOfQueuedFacts = 0;
                }
            }
            catch (JRDFoxException e) {
                throw new RuntimeException(e);
            }
        }
        
        public void endProcessing() {
            try {
                if (m_facts.length() != 0)
                	m_dataStore.importText(m_facts.toString());
                if (m_rules.size() != 0)
                	m_dataStore.importRules(m_rules);
            }
            catch (JRDFoxException e) {
                throw new RuntimeException(e);
            }
            finally {
                m_rules.clear();
                m_facts.delete(0, m_facts.length());
                m_numberOfQueuedFacts = 0;
            }
        }
    }
    
    protected static final Individual ABOUT_NULLARY = Individual.create("about.nullary");

}
