package uk.ac.ox.cs.goalDriven.RDFoxBridge;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Arrays;
import java.util.Collections;
import java.util.function.Function;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.parser.CSVParser;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class JRDFoxBridgeExample {

    public static Reader getResource(String resourceName) throws IOException {
        InputStream resourceStream = JRDFoxBridgeExample.class.getResourceAsStream(resourceName);
        if (resourceStream == null)
            throw new IOException("Cannot locate resource '" + resourceName + "'.");
        return new InputStreamReader(resourceStream);
    }

    public static void main(String[] args) throws Exception {
        Predicate N = Predicate.create("N");
        Predicate U = Predicate.create("U");
        Predicate D = Predicate.create("D");
        Predicate B1 = Predicate.create("B1");
        Predicate B2 = Predicate.create("B2");
        Predicate R = Predicate.create("R");
        Predicate T = Predicate.create("T");

        Variable X = Variable.create("X");
        Variable Y = Variable.create("Y");
        Variable Z = Variable.create("Z");

        // Create the bridge to JRDFox. The argument determines whether we are using
        // exact equality.
        // The object must be deleted after use, which is why we used a
        // try-with-resources statement.
        try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(false)) {
            // The input processor allows adding rules and data to the data store wrapped by
            // the bridge.
            InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();
            // We are now ready to import the data. The code should be self-explanatory,
            // but note that any importation should be enclosed in startProcessing() --
            // endProcessing().
            inputProcessor.startProcessing();
            inputProcessor.processFact(N, Collections.emptyList());
            try (Reader reader = getResource("res/U.csv")) {
                CSVParser.parse(reader, U, inputProcessor);
            }
            try (Reader reader = getResource("res/B1.csv")) {
                CSVParser.parse(reader, B1, inputProcessor);
            }
            try (Reader reader = getResource("res/B2.csv")) {
                CSVParser.parse(reader, B2, inputProcessor);
            }
            try (Reader reader = getResource("res/T.csv")) {
                CSVParser.parse(reader, Predicate.create("T"), inputProcessor);
            }
            inputProcessor.endProcessing();
            // We can now import the rules. Note that startProcessing() -- endProcessing()
            // is in this case called from within the parser.
            try (Reader reader = getResource("res/rules.txt")) {
                new CommonParser(reader).parse(inputProcessor);
            }
            // We compute the materialization.
            jrdfoxBridge.computeMaterialization();

            // We can export the content of the data store to disk in Datalog format.
            jrdfoxBridge.getDataStore().export(new File("chaseBenchDemo.txt"), "Turtle");

            // We can now evaluate queries over the computed facts. This query provides two
            // ways to do this.

            // We can retrieve "raw resources" as they come out of RDFox. This is faster as
            // it does not involve any transformation of results.
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X, Y),
                    Arrays.asList(Atom.create(N), Atom.create(U, X), Atom.create(B1, X, Y)),
                    Function.identity(), false, true);
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X, Y, Z),
                    Arrays.asList(Atom.create(T, X, Y, Z)),
                    Function.identity(), false, true);
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X),
                    Arrays.asList(Atom.create(D, X)),
                    Function.identity(), false, true);
            // However, function symbols are returned as RDFox blank nodes, as shown in the
            // following query.
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X, Y),
                    Arrays.asList(Atom.create(R, X, Y)),
                    Function.identity(), false, true);
            // We can also filter out labelled nulls in the answers and thus get just the
            // certain answers.
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X, Y),
                    Arrays.asList(Atom.create(R, X, Y)),
                    Function.identity(), true, true);
            // We can also filter out labelled nulls in the answers and thus get just the
            // certain answers.
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X),
                    Arrays.asList(Atom.create(R, X, Y)),
                    Function.identity(), true, true);

            // The JRDFox bridge also provides a way to transform these blank nodes back
            // into function symbols.
            JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, Arrays.asList(X, Y), Arrays.asList(Atom.create(R, X, Y)),
                    (x) -> {
                        return jrdfoxBridge.parseResource(x);
                    },
                    false, true);

            System.out.println();
        }
    }

}
