package uk.ac.ox.cs.goalDriven.model;

import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Collection;
import java.util.Map;

public class FunctionTerm extends Term implements Serializable {
    private static final long serialVersionUID = 5950450463145714290L;

    protected final Function m_function;
    protected final Term[] m_arguments;

    protected FunctionTerm(Function function, Term[] arguments) {
        this.m_function = function;
        this.m_arguments = arguments.clone();
    }

    public Function getFunction() {
        return m_function;
    }

    public int getNumberOfArguments() {
        return m_arguments.length;
    }

    public Term getArgument(int index) {
        return m_arguments[index];
    }

    public Term[] getArguments() {
        return m_arguments.clone();
    }

    public void getArguments(Term[] arguments) {
        System.arraycopy(m_arguments, 0, arguments, 0, m_arguments.length);
    }

    public void getFreeVariables(Collection<Variable> collection) {
        for (Term argument : m_arguments)
            argument.getFreeVariables(collection);
    }

    public void getConstants(Collection<Constant> collection) {
        for (Term argument : m_arguments)
            argument.getConstants(collection);
    }

    public void getFunctions(Map<Function,Integer> collection) {
        collection.put(m_function, m_arguments.length);
        for (Term argument : m_arguments)
            argument.getFunctions(collection);
    }

    public boolean isGround() {
        for (int index = m_arguments.length - 1; index >= 0; --index)
            if (!m_arguments[index].isGround())
                return false;
        return true;
    }

    public FunctionTerm replaceFunction(Function newFunction) {
        return create(newFunction, m_arguments);
    }

    public FunctionTerm applySubstitution(Map<Variable, ? extends Term> substitution) {
        if (substitution.isEmpty())
            return this;
        else {
            Term[] arguments = new Term[m_arguments.length];
            for (int index = 0; index < m_arguments.length; index++)
                arguments[index] = m_arguments[index].applySubstitution(substitution);
            return FunctionTerm.create(m_function, arguments);
        }
    }

    public FunctionTerm applyMapping(Map<? extends Term, ? extends Term> mapping) {
        if (mapping.isEmpty())
            return this;
        else {
            Term[] arguments = new Term[m_arguments.length];
            for (int index = 0; index < m_arguments.length; index++)
                arguments[index] = m_arguments[index].applyMapping(mapping);
            return FunctionTerm.create(m_function, arguments);
        }
    }

    public void toString(StringBuilder builder) {
        m_function.toString(builder);
        if (m_arguments.length > 0) {
            builder.append('(');
            for (int index = 0; index < m_arguments.length; index++) {
                if (index > 0)
                    builder.append(',');
                m_arguments[index].toString(builder);
            }
            builder.append(')');
        }
    }

    public void print(PrintWriter output) {
        m_function.print(output);
        if (m_arguments.length > 0) {
            output.print('(');
            for (int index = 0; index < m_arguments.length; index++) {
                if (index > 0)
                    output.print(',');
                m_arguments[index].print(output);
            }
            output.print(')');
        }
    }

    protected Object readResolve() {
        return s_interningManager.intern(this);
    }

    protected static final InterningManager<FunctionTerm> s_interningManager = new InterningManager<FunctionTerm>() {
        protected boolean equal(FunctionTerm object1, FunctionTerm object2) {
            if (!object1.m_function.equals(object2.m_function)
                    || object1.m_arguments.length != object2.m_arguments.length)
                return false;
            for (int index = object1.m_arguments.length - 1; index >= 0; --index)
                if (!object1.m_arguments[index].equals(object2.m_arguments[index]))
                    return false;
            return true;
        }

        protected int getHashCode(FunctionTerm object) {
            int hashCode = object.m_function.hashCode();
            for (int index = object.m_arguments.length - 1; index >= 0; --index)
                hashCode = hashCode * 7 + object.m_arguments[index].hashCode();
            return hashCode;
        }
    };

    public static FunctionTerm create(Function function, Term... arguments) {
        return s_interningManager.intern(new FunctionTerm(function, arguments));
    }

    @Override
    public int depth() {
        int maxDepth = 0;
        for (int argumentIndex = 0; argumentIndex < m_arguments.length; ++argumentIndex) {
            int d = m_arguments[argumentIndex].depth();
            if (maxDepth < d)
                maxDepth = d;
        }
        return maxDepth + 1;
    }
}
