package uk.ac.ox.cs.goalDriven.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.List;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.LabeledNull;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class CommonParser {
    protected final StreamTokenizer m_tokenizer;

    public CommonParser(Reader reader) throws IOException {
        m_tokenizer = new StreamTokenizer(reader);
        m_tokenizer.resetSyntax();
        m_tokenizer.commentChar('#');
        m_tokenizer.eolIsSignificant(false);
        m_tokenizer.whitespaceChars(' ', ' ');
        m_tokenizer.whitespaceChars('\r', '\r');
        m_tokenizer.whitespaceChars('\n', '\n');
        m_tokenizer.whitespaceChars('\t', '\t');
        m_tokenizer.wordChars('_', '_');
        m_tokenizer.wordChars('a', 'z');
        m_tokenizer.wordChars('A', 'Z');
        m_tokenizer.wordChars('0', '9');
        m_tokenizer.wordChars('+', '+');
        m_tokenizer.wordChars('-', '-');
        m_tokenizer.wordChars('?', '?');
        m_tokenizer.wordChars('<', '<');
        m_tokenizer.wordChars('>', '>');
        m_tokenizer.quoteChar('"');
        m_tokenizer.nextToken();
    }

    public Term parseTerm() throws IOException {
    	if (m_tokenizer.ttype == '"') {
    		Constant constant = Constant.create('"' + m_tokenizer.sval + '"');
    		m_tokenizer.nextToken();
    		return constant;
    	}
    	else if (m_tokenizer.ttype == StreamTokenizer.TT_WORD) {
    		String symbol = m_tokenizer.sval;
    		m_tokenizer.nextToken();
    		if (symbol.startsWith("?"))
    			return Variable.create(symbol.substring(1));
    		else if (symbol.startsWith("_:"))
    			return LabeledNull.create(symbol.substring(2));
    		else if (m_tokenizer.ttype == '(') {
    			m_tokenizer.nextToken();
    			List<Term> arguments = new ArrayList<Term>();
    			arguments.add(parseTerm());
    			while (m_tokenizer.ttype == ',') {
    				m_tokenizer.nextToken();
    				arguments.add(parseTerm());
    			}
    			if (m_tokenizer.ttype != ')')
    	            throw new IOException("Illegal functional term.");
    			m_tokenizer.nextToken();
    			return FunctionTerm.create(Function.create(symbol), arguments.toArray(new Term[arguments.size()]));
    		}
    		else
    			return Constant.create(symbol);
    	}
    	else
            throw new IOException("Illegal term.");
    }
    
    public Atom parseAtom() throws IOException {
        Predicate predicate;
        if (m_tokenizer.ttype == StreamTokenizer.TT_WORD)
            predicate = Predicate.create(m_tokenizer.sval);
        else if (m_tokenizer.ttype == '=')
            predicate = Predicate.EQUALS;
        else
            throw new IOException("Illegal predicate type.");
        m_tokenizer.nextToken();
        List<Term> arguments = new ArrayList<Term>();
        if (m_tokenizer.ttype == '(') {
        	m_tokenizer.nextToken();
        	arguments.add(parseTerm());
        	while (m_tokenizer.ttype == ',') {
        		m_tokenizer.nextToken();
            	arguments.add(parseTerm());
        	}
			if (m_tokenizer.ttype != ')')
	            throw new IOException("Illegal functional atom.");
			m_tokenizer.nextToken();
        }
        if (predicate.equals(Predicate.EQUALS) && arguments.size() != 2)
            throw new IOException("Illegal arguments: = accepts two arguments only.");
        return Atom.create(predicate, arguments.toArray(new Term[arguments.size()]));
    }

    public Object parseRuleOrAtom() throws IOException {
        List<Atom> before = new ArrayList<Atom>();
        if (m_tokenizer.ttype != StreamTokenizer.TT_WORD || (!"<-".equals(m_tokenizer.sval) && !"->".equals(m_tokenizer.sval))) {
            Atom atom = parseAtom();
            if (m_tokenizer.ttype == '.') {
                m_tokenizer.nextToken();
                return atom;
            }
            before.add(atom);
            while (m_tokenizer.ttype == ',' || m_tokenizer.ttype == '&') {
                m_tokenizer.nextToken();
                before.add(parseAtom());
            }
        }
        boolean leftToRight;
        if (m_tokenizer.ttype == StreamTokenizer.TT_WORD && "<-".equals(m_tokenizer.sval))
            leftToRight = false;
        else if (m_tokenizer.ttype == StreamTokenizer.TT_WORD && "->".equals(m_tokenizer.sval))
            leftToRight = true;
        else
            throw new IOException("Expected '<-' or '->'.");
        m_tokenizer.nextToken();
        List<Atom> after = new ArrayList<Atom>();
        if (m_tokenizer.ttype != '.') {
            after.add(parseAtom());
            while (m_tokenizer.ttype == ',' || m_tokenizer.ttype == '&') {
                m_tokenizer.nextToken();
                after.add(parseAtom());
            }
        }
        if (m_tokenizer.ttype != '.')
            throw new IOException("Expected '.' at the end of the rule.");
        m_tokenizer.nextToken();
        Atom[] beforeAtoms = before.toArray(new Atom[before.size()]);
        Atom[] afterAtoms = after.toArray(new Atom[after.size()]);
        if (leftToRight)
            return Rule.create(afterAtoms, beforeAtoms);
        else
            return Rule.create(beforeAtoms, afterAtoms);
    }

    public void parse(InputProcessor inputProcessor) throws IOException {
        inputProcessor.startProcessing();
        try {
            while (m_tokenizer.ttype != StreamTokenizer.TT_EOF) {
                Object ruleOrAtom = parseRuleOrAtom();
                if (ruleOrAtom instanceof Rule)
                    inputProcessor.processRule((Rule) ruleOrAtom);
                else if (ruleOrAtom instanceof Atom) {
                    Atom atom = (Atom) ruleOrAtom;
                    List<String> argumentLexicalForms = new ArrayList<String>();
                    for (int index = 0; index < atom.getNumberOfArguments(); ++index) {
                        Term argument = atom.getArgument(index);
                        if (argument instanceof Constant)
                            argumentLexicalForms.add(((Constant) argument).getLexicalForm());
                        else
                            throw new IOException("Facts should only contain constants and labeled nulls as arguments.");
                    }
                    inputProcessor.processFact(atom.getPredicate(), argumentLexicalForms);
                }
            }
        }
        finally {
            inputProcessor.endProcessing();
        }
    }

    public static void parse(File file, InputProcessor inputProcessor) throws IOException {
        Reader input = new InputStreamReader(new FileInputStream(file));
        try {
            new CommonParser(input).parse(inputProcessor);
        }
        finally {
            input.close();
        }
    }
}
