package uk.ac.ox.cs.goalDriven.processors;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.LabeledNull;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

public abstract class Skolemization implements InputProcessor {
    protected final boolean m_breakIndependentRuleHeads;
    protected final boolean m_useArguments;
    protected final InputProcessor m_inputProcessor;
    protected final boolean m_discardRulesWithNullsInTheHead;

    public Skolemization(boolean breakIndependentRuleHeads, boolean useArguments,
            boolean discardRulesWithNullsInTheHead, InputProcessor inputProcessor) {
        m_breakIndependentRuleHeads = breakIndependentRuleHeads;
        m_inputProcessor = inputProcessor;
        m_useArguments = useArguments;
        m_discardRulesWithNullsInTheHead = discardRulesWithNullsInTheHead;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
    }

    public void processRule(Rule rule) {
        // Compute the body variables.
        Set<Variable> bodyVariables = new LinkedHashSet<Variable>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex)
            rule.getBodyAtom(bodyAtomIndex).getFreeVariables(bodyVariables);
        // Determine the head variables and build the reachability graph.
        ReachabilityGraph reachabilityGraph = new ReachabilityGraph(bodyVariables);
        Set<Variable> headVariables = new LinkedHashSet<Variable>();
        for (int headAtomIndex = 0; headAtomIndex < rule.getNumberOfHeadAtoms(); ++headAtomIndex) {
            Atom headAtom = rule.getHeadAtom(headAtomIndex);
            headAtom.getFreeVariables(headVariables);
            reachabilityGraph.addAtom(headAtom);
        }
        // Break the heads if necessary.
        if (m_breakIndependentRuleHeads) {
            Set<Atom> processedHeadAtoms = new HashSet<Atom>();
            for (int headIndex = 0; headIndex < rule.getNumberOfHeadAtoms(); ++headIndex) {
                Atom headAtom = rule.getHeadAtom(headIndex);
                if (!processedHeadAtoms.contains(headAtom)) {
                    Set<Variable> existentialVariables = new LinkedHashSet<Variable>();
                    List<Atom> newHead = new ArrayList<Atom>();
                    reachabilityGraph.appendReachableAtoms(headAtom, newHead, processedHeadAtoms, existentialVariables);
                    Map<Variable, Term> substitution = new LinkedHashMap<Variable, Term>();
                    for (Variable existentialVariable : existentialVariables)
                        substitution.put(existentialVariable, getSkolemTerm(existentialVariable, reachabilityGraph));
                    for (int newHeadIndex = 0; newHeadIndex < newHead.size(); ++newHeadIndex)
                        newHead.set(newHeadIndex, newHead.get(newHeadIndex).applySubstitution(substitution));
                    Atom[] newBody = rule.getBodyAtoms();
                    for (int newBodyIndex = 0; newBodyIndex < newBody.length; ++newBodyIndex)
                        newBody[newBodyIndex].applySubstitution(substitution);
                    for (int newHeadIndex = 0; newHeadIndex < newHead.size(); ++newHeadIndex) {
                        Rule newRule = Rule.create(newHead.get(newHeadIndex), newBody);
                        if (!m_discardRulesWithNullsInTheHead)
                            m_inputProcessor.processRule(newRule);
                        else {
                            boolean addRule = true;
                            Atom head = newRule.getHeadAtom(0);
                            for (int argumentIndex = 0; argumentIndex < head.getNumberOfArguments(); ++argumentIndex) {
                                if (head.getArgument(argumentIndex) instanceof LabeledNull)
                                    addRule = false;
                            }
                            if (addRule)
                                m_inputProcessor.processRule(newRule);
                        }

                    }
                }
            }
        } else {
            Map<Variable, Term> substitution = new LinkedHashMap<Variable, Term>();
            for (Variable existentialVariable : reachabilityGraph.getExistentialVariables())
                substitution.put(existentialVariable, getSkolemTerm(existentialVariable, reachabilityGraph));
            m_inputProcessor.processRule(rule.applySubstitution(substitution));
        }
    }

    protected Term getSkolemTerm(Variable existentialVariable, ReachabilityGraph reachabilityGraph) {
        if (!m_useArguments)
            return LabeledNull.create("ex" + getNextExistentialCounter());
        List<Term> arguments = new ArrayList<Term>();
        reachabilityGraph.appendReachableBodyVariables(existentialVariable, arguments);
        if (arguments.isEmpty())
            return LabeledNull.create("ex" + getNextExistentialCounter());
        else
            return FunctionTerm.create(Function.create("ex" + getNextExistentialCounter()),
                    arguments.toArray(new Term[arguments.size()]));
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        m_inputProcessor.endProcessing();
    }

    protected abstract int getNextExistentialCounter();

    protected static class ReachabilityGraph {
        protected final Set<Variable> m_bodyVariables;
        protected final Map<Variable, Set<Atom>> m_atomsByExistentialVariables;

        public ReachabilityGraph(Set<Variable> bodyVariables) {
            m_bodyVariables = bodyVariables;
            m_atomsByExistentialVariables = new LinkedHashMap<Variable, Set<Atom>>();
        }

        public void addAtom(Atom atom) {
            for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
                Term argument = atom.getArgument(argumentIndex);
                if (argument instanceof Variable && !m_bodyVariables.contains(argument)) {
                    Variable variable = (Variable) argument;
                    Set<Atom> atoms = m_atomsByExistentialVariables.get(variable);
                    if (atoms == null) {
                        atoms = new LinkedHashSet<Atom>();
                        m_atomsByExistentialVariables.put(variable, atoms);
                    }
                    atoms.add(atom);
                }
            }
        }

        public Set<Variable> getExistentialVariables() {
            return m_atomsByExistentialVariables.keySet();
        }

        public void appendReachableAtoms(Atom start, List<Atom> result, Set<Atom> inResult,
                Set<Variable> existentialVariables) {
            List<Atom> atomQueue = new ArrayList<Atom>();
            Set<Atom> inAtomQueue = new HashSet<Atom>();
            atomQueue.add(start);
            inAtomQueue.add(start);
            int index = 0;
            while (index < atomQueue.size()) {
                Atom current = atomQueue.get(index++);
                result.add(current);
                inResult.add(current);
                for (int argumentIndex = 0; argumentIndex < current.getNumberOfArguments(); ++argumentIndex) {
                    Term argument = current.getArgument(argumentIndex);
                    Set<Atom> atoms = m_atomsByExistentialVariables.get(argument);
                    if (atoms != null) {
                        existentialVariables.add((Variable) argument);
                        for (Atom atom : atoms)
                            if (inAtomQueue.add(atom))
                                atomQueue.add(atom);
                    }
                }
            }
        }

        public void appendReachableBodyVariables(Variable start, List<Term> result) {
            Set<Term> inResult = new HashSet<Term>();
            List<Variable> variableQueue = new ArrayList<Variable>();
            Set<Variable> inVariableQueue = new HashSet<Variable>();
            variableQueue.add(start);
            inVariableQueue.add(start);
            int index = 0;
            while (index < variableQueue.size()) {
                Variable current = variableQueue.get(index++);
                for (Atom atom : m_atomsByExistentialVariables.get(current)) {
                    for (int argumentIndex = 0; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
                        Term argument = atom.getArgument(argumentIndex);
                        if (argument instanceof Variable) {
                            Variable variable = (Variable) argument;
                            if (m_bodyVariables.contains(variable)) {
                                if (inResult.add(variable))
                                    result.add(variable);
                            } else {
                                if (inVariableQueue.add(variable))
                                    variableQueue.add(variable);
                            }
                        }
                    }
                }
            }
        }
    }
}
